# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Compare DS9 parsing of the astropy regions package to pyregion.

This scripts compares the DS9 parsing of the astropy regions package to
pyregion in two regards.

* speed : the time to parse a file given ``REPETITIONS`` repetitions

* completeness : the ratio of regions in a file (estimated by the number
  of '(' in the file) that survives the parsing process

The test files are the ones used since PyAstro16 (see
https://zenodo.org/record/56793).
"""

import re
import timeit
from pathlib import Path

import numpy as np
import pyregion
from astropy.table import Table

from regions import DS9RegionParserError, read_ds9

TEST_FILE_DIR = Path('../regions/io/ds9/tests/data')
REPETITIONS = 1

p_region_count = re.compile(r'[^=\)]\(')

results = list()

filenames = sorted(TEST_FILE_DIR.glob('*.reg'))

keep_filenames = []
for filename in filenames:
    skipfiles = ('ds9.linear.wcs.reg', 'wcsa', 'wcsc', 'wcsd', 'wcsi',
                 'wcsp')
    skip = False
    for skipfile in skipfiles:
        if skipfile in filename.name:
            print(f'Skipping {filename}')
            skip = True
            continue
    if not skip:
        keep_filenames.append(filename)

for filename in keep_filenames:
    print(f'\n\n{filename}')

    # estimate total number of regions
    n_regions = 0
    with open(str(filename)) as origin_file:
        for line in origin_file:
            n_regions += len(p_region_count.findall(line))

    # regions
    try:
        region_regions = read_ds9(str(filename), errors='warn')
    except DS9RegionParserError:
        time_regions = -1
        compl_regions = 0
    else:
        time_regions = timeit.timeit("read_ds9(str(filename), errors='warn')",
                                     setup='from regions import read_ds9',
                                     globals=globals(),
                                     number=REPETITIONS) / REPETITIONS

        compl_regions = np.divide(len(region_regions), n_regions)

    # pyregion
    try:
        pyregion_regions = pyregion.open(str(filename))
    except ValueError:
        time_pyregion = -1
        compl_pyregion = 0
    else:
        time_pyregion = timeit.timeit('pyregion.open(str(filename))',
                                      setup='import pyregion',
                                      globals=globals(),
                                      number=REPETITIONS) / REPETITIONS

        compl_pyregion = np.divide(len(pyregion_regions), n_regions)

    # collect results
    results.append(dict(filename=filename.parts[-1],
                        time_regions=time_regions,
                        time_pyregion=time_pyregion,
                        compl_regions=compl_regions,
                        compl_pyregion=compl_pyregion))


result_table = Table()
result_table['filename'] = [_['filename'] for _ in results]
result_table['time_regions'] = [_['time_regions'] for _ in results]
result_table['time_regions'].format = '.3f'
result_table['time_pyregion'] = [_['time_pyregion'] for _ in results]
result_table['time_pyregion'].format = '.3f'
result_table['compl_pyregion'] = [_['compl_pyregion'] for _ in results]
result_table['compl_pyregion'].format = '0.1%'
result_table['compl_regions'] = [_['compl_regions'] for _ in results]
result_table['compl_regions'].format = '0.1%'
result_table.meta['repetitions for timing'] = REPETITIONS

savename = 'regions_pyregion_comparison.csv'
print(f'Writing {savename}')
result_table.write(savename)
result_table.pprint()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines elliptical regions in both pixel and sky
coordinates.
"""

import math

import astropy.units as u
import numpy as np
from astropy.coordinates import Angle

from regions._geometry import elliptical_overlap_grid
from regions._utils.wcs_helpers import pixel_scale_angle_at_skycoord
from regions.core.attributes import (PositiveScalar, PositiveScalarAngle,
                                     RegionMetaDescr, RegionVisualDescr,
                                     ScalarAngle, ScalarPixCoord,
                                     ScalarSkyCoord)
from regions.core.bounding_box import RegionBoundingBox
from regions.core.core import PixelRegion, SkyRegion
from regions.core.mask import RegionMask
from regions.core.metadata import RegionMeta, RegionVisual
from regions.core.pixcoord import PixCoord

__all__ = ['EllipsePixelRegion', 'EllipseSkyRegion']


class EllipsePixelRegion(PixelRegion):
    """
    An ellipse in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the ellipse.
    width : float
        The width of the ellipse (before rotation) in pixels.
    height : float
        The height of the ellipse (before rotation) in pixels.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the ellipse, measured anti-clockwise. If
        set to zero (the default), the width axis is lined up with the x
        axis.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from astropy.coordinates import Angle
        from regions import PixCoord, EllipsePixelRegion
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1)

        reg = EllipsePixelRegion(PixCoord(15, 10), width=16, height=10,
                                 angle=Angle(30, 'deg'))
        patch = reg.plot(ax=ax, facecolor='none', edgecolor='red', lw=2,
                         label='Ellipse')

        ax.legend(handles=(patch,), loc='upper center')
        ax.set_xlim(0, 30)
        ax.set_ylim(0, 20)
        ax.set_aspect('equal')
    """

    _params = ('center', 'width', 'height', 'angle')
    _mpl_artist = 'Patch'
    center = ScalarPixCoord('The center pixel position as a |PixCoord|.')
    width = PositiveScalar('The width of the ellipse (before rotation) in '
                           'pixels as a float.')
    height = PositiveScalar('The height of the ellipse (before rotation) in '
                            'pixels as a float.')
    angle = ScalarAngle('The rotation angle measured anti-clockwise as a '
                        '|Quantity| angle.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, center, width, height, angle=0. * u.deg, meta=None,
                 visual=None):
        self.center = center
        self.width = width
        self.height = height
        self.angle = angle
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    @property
    def area(self):
        return math.pi / 4 * self.width * self.height

    def contains(self, pixcoord):
        pixcoord = PixCoord._validate(pixcoord, name='pixcoord')
        cos_angle = np.cos(self.angle)
        sin_angle = np.sin(self.angle)
        dx = pixcoord.x - self.center.x
        dy = pixcoord.y - self.center.y
        in_ell = ((2 * (cos_angle * dx + sin_angle * dy) / self.width) ** 2
                  + (2 * (sin_angle * dx - cos_angle * dy)
                     / self.height) ** 2 <= 1.)
        if self.meta.get('include', True):
            return in_ell
        else:
            return np.logical_not(in_ell)

    def to_sky(self, wcs):
        center = wcs.pixel_to_world(self.center.x, self.center.y)
        _, pixscale, north_angle = pixel_scale_angle_at_skycoord(center, wcs)
        height = Angle(self.height * u.pix * pixscale, 'arcsec')
        width = Angle(self.width * u.pix * pixscale, 'arcsec')
        # region sky angles are defined relative to the WCS longitude axis;
        # photutils aperture sky angles are defined as the PA of the
        # semimajor axis (i.e., relative to the WCS latitude axis)
        angle = self.angle - (north_angle - 90 * u.deg)
        return EllipseSkyRegion(center, width, height, angle=angle,
                                meta=self.meta.copy(),
                                visual=self.visual.copy())

    @property
    def bounding_box(self):
        """
        The minimal bounding box (`~regions.RegionBoundingBox`)
        enclosing the exact elliptical region.
        """
        # We use the solution described in
        # https://stackoverflow.com/a/14163413
        cos_theta = np.cos(self.angle)
        sin_theta = np.sin(self.angle)
        width_x = 0.5 * self.width * cos_theta
        width_y = 0.5 * self.width * sin_theta
        height_x = 0.5 * self.height * -sin_theta
        height_y = 0.5 * self.height * cos_theta
        dx = np.sqrt(width_x**2 + height_x**2)
        dy = np.sqrt(width_y**2 + height_y**2)

        xmin = self.center.x - dx
        xmax = self.center.x + dx
        ymin = self.center.y - dy
        ymax = self.center.y + dy

        return RegionBoundingBox.from_float(xmin, xmax, ymin, ymax)

    def to_mask(self, mode='center', subpixels=5):
        # NOTE: assumes this class represents a single circle

        self._validate_mode(mode, subpixels)

        if mode == 'center':
            mode = 'subpixels'
            subpixels = 1

        # Find bounding box and mask size
        bbox = self.bounding_box
        ny, nx = bbox.shape

        # Find position of pixel edges and recenter so that ellipse is
        # at origin
        xmin = float(bbox.ixmin) - 0.5 - self.center.x
        xmax = float(bbox.ixmax) - 0.5 - self.center.x
        ymin = float(bbox.iymin) - 0.5 - self.center.y
        ymax = float(bbox.iymax) - 0.5 - self.center.y

        use_exact = 0 if mode == 'subpixels' else 1

        fraction = elliptical_overlap_grid(xmin, xmax, ymin, ymax, nx, ny,
                                           0.5 * self.width, 0.5 * self.height,
                                           self.angle.to(u.rad).value,
                                           use_exact, subpixels)

        return RegionMask(fraction, bbox=bbox)

    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Return a matplotlib patch object for the region
        (`matplotlib.patches.Ellipse`).

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        **kwargs : dict
            Any keyword arguments accepted by
            `~matplotlib.patches.Ellipse`. These keywords will override
            any visual meta attributes of this region.

        Returns
        -------
        artist : `~matplotlib.patches.Ellipse`
            A matplotlib ellipse patch.
        """
        from matplotlib.patches import Ellipse

        xy = self.center.x - origin[0], self.center.y - origin[1]
        width = self.width
        height = self.height
        # matplotlib expects rotation in degrees (anti-clockwise)
        angle = self.angle.to('deg').value

        mpl_kwargs = self.visual.define_mpl_kwargs(self._mpl_artist)
        mpl_kwargs.update(kwargs)

        return Ellipse(xy=xy, width=width, height=height, angle=angle,
                       **mpl_kwargs)

    def _update_from_mpl_selector(self, *args, **kwargs):
        xmin, xmax, ymin, ymax = self._mpl_selector.extents
        self.center = PixCoord(x=0.5 * (xmin + xmax),
                               y=0.5 * (ymin + ymax))
        self.width = (xmax - xmin)
        self.height = (ymax - ymin)
        self.angle = 0. * u.deg
        if self._mpl_selector_callback is not None:
            self._mpl_selector_callback(self)

    def as_mpl_selector(self, ax, active=True, sync=True, callback=None,
                        drag_from_anywhere=False, **kwargs):
        """
        Return a matplotlib editable widget for this region
        (`matplotlib.widgets.EllipseSelector`).

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`
            The matplotlib axes to add the selector to.
        active : bool, optional
            Whether the selector should be active by default.
        sync : bool, optional
            If `True` (the default), the region will be kept in
            sync with the selector. Otherwise, the selector will be
            initialized with the values from the region but the two will
            then be disconnected.
        callback : callable, optional
            If specified, this function will be called every time the
            region is updated. This only has an effect if ``sync`` is
            `True`. If a callback is set, it is called for the first
            time once the selector has been created.
        drag_from_anywhere : bool, optional
            If `True`, the selector can be moved by clicking anywhere within
            its bounds, else only at the central anchor
            (only available with matplotlib 3.5 upwards; default: `False`).
        **kwargs : dict
            Additional keyword arguments that are passed to
            `matplotlib.widgets.EllipseSelector`.

        Returns
        -------
        selector : `matplotlib.widgets.EllipseSelector`
            The matplotlib selector.

        Notes
        -----
        Once a selector has been created, you will need to keep a
        reference to it until you no longer need it. In addition,
        you can enable/disable the selector at any point by calling
        ``selector.set_active(True)`` or ``selector.set_active(False)``.
        """
        from matplotlib.widgets import EllipseSelector

        if hasattr(self, '_mpl_selector'):
            raise AttributeError('Cannot attach more than one selector to a region.')

        if self.angle.value != 0:
            raise NotImplementedError('Cannot create matplotlib selector for rotated ellipse.')

        if sync:
            sync_callback = self._update_from_mpl_selector
        else:
            def sync_callback(*args, **kwargs):
                pass

        rectprops = {'edgecolor': self.visual.get('color', 'black'),
                     'facecolor': 'none',
                     'linewidth': self.visual.get('linewidth', 1),
                     'linestyle': self.visual.get('linestyle', 'solid')}
        rectprops.update(kwargs.pop('props', dict()))
        kwargs.update({'props': rectprops})

        self._mpl_selector = EllipseSelector(
            ax, sync_callback, interactive=True,
            drag_from_anywhere=drag_from_anywhere, **kwargs)

        self._mpl_selector.extents = (self.center.x - self.width / 2,
                                      self.center.x + self.width / 2,
                                      self.center.y - self.height / 2,
                                      self.center.y + self.height / 2)
        self._mpl_selector.set_active(active)
        self._mpl_selector_callback = callback

        if sync and self._mpl_selector_callback is not None:
            self._mpl_selector_callback(self)

        return self._mpl_selector

    def rotate(self, center, angle):
        """
        Rotate the region.

        Positive ``angle`` corresponds to counter-clockwise rotation.

        Parameters
        ----------
        center : `~regions.PixCoord`
            The rotation center point.
        angle : `~astropy.coordinates.Angle`
            The rotation angle.

        Returns
        -------
        region : `EllipsePixelRegion`
            The rotated region (which is an independent copy).
        """
        center = self.center.rotate(center, angle)
        angle = self.angle + angle
        return self.copy(center=center, angle=angle)


class EllipseSkyRegion(SkyRegion):
    """
    An ellipse defined using sky coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the center of the ellipse.
    width : `~astropy.units.Quantity`
        The width of the ellipse (before rotation) as an angle.
    height : `~astropy.units.Quantity`
        The height of the ellipse (before rotation) as an angle.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the ellipse, measured anti-clockwise. If
        set to zero (the default), the width axis is lined up with the
        longitude axis of the celestial coordinates.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.
    """

    _params = ('center', 'width', 'height', 'angle')
    center = ScalarSkyCoord('The center position as a |SkyCoord|.')
    width = PositiveScalarAngle('The width of the ellipse (before rotation) '
                                'as a |Quantity| angle.')
    height = PositiveScalarAngle('The height of the ellipse (before rotation) '
                                 'as a |Quantity| angle.')
    angle = ScalarAngle('The rotation angle measured anti-clockwise as a '
                        '|Quantity| angle.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, center, width, height, angle=0. * u.deg, meta=None,
                 visual=None):
        self.center = center
        self.width = width
        self.height = height
        self.angle = angle
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    def to_pixel(self, wcs):
        center, pixscale, north_angle = pixel_scale_angle_at_skycoord(
            self.center, wcs)
        height = (self.height / pixscale).to(u.pixel).value
        width = (self.width / pixscale).to(u.pixel).value
        # region sky angles are defined relative to the WCS longitude axis;
        # photutils aperture sky angles are defined as the PA of the
        # semimajor axis (i.e., relative to the WCS latitude axis)
        angle = self.angle + (north_angle - 90 * u.deg)
        return EllipsePixelRegion(center, width, height, angle=angle,
                                  meta=self.meta.copy(),
                                  visual=self.visual.copy())



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from regions.shapes import (CircleAnnulusPixelRegion, CirclePixelRegion,
                            EllipseAnnulusPixelRegion, EllipsePixelRegion,
                            PointPixelRegion, PolygonPixelRegion,
                            RectanglePixelRegion)

__all__ = []


# mappings from FITS shape to region class and column names/indices
shape_map = {'point': (PointPixelRegion, ('X0', 'Y0')),
             'circle': (CirclePixelRegion, ('X0', 'Y0', 'R0')),
             'ellipse': (EllipsePixelRegion,
                         ('X0', 'Y0', 'R0', 'R1', 'ROTANG0')),
             'annulus': (CircleAnnulusPixelRegion, ('X0', 'Y0', 'R0', 'R1')),
             'elliptannulus': (EllipseAnnulusPixelRegion,
                               ('X0', 'Y0', 'R0', 'R1', 'R2', 'R3',
                                'ROTANG0')),
             'box': (RectanglePixelRegion, ('X0', 'Y0', 'R0', 'R1')),
             'rotbox': (RectanglePixelRegion,
                        ('X0', 'Y0', 'R0', 'R1', 'ROTANG0')),
             'rectangle': (RectanglePixelRegion, ('X0', 'X1', 'Y0', 'Y1')),
             'rotrectangle': (RectanglePixelRegion,
                              ('X0', 'X1', 'Y0', 'Y1', 'ROTANG0')),
             'polygon': (PolygonPixelRegion, ('X', 'Y'))}


class FITSParserError(Exception):
    """
    A custom exception for FITS parsing errors.
    """



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import re
import string
import warnings
from dataclasses import dataclass

import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.utils.data import get_readable_fileobj
from astropy.utils.exceptions import AstropyUserWarning

from regions.core import PixCoord, RegionMeta, Regions, RegionVisual
from regions.core.registry import RegionsRegistry
from regions.io.ds9.core import (DS9ParserError, ds9_frame_map,
                                 ds9_params_template, ds9_shape_to_region,
                                 make_region_template)
from regions.io.ds9.meta import _split_raw_metadata, _translate_ds9_to_visual

__all__ = []


@RegionsRegistry.register(Regions, 'read', 'ds9')
def _read_ds9(filename, cache=False):
    """
    Read a DS9 region file in as a list of `~regions.Region` objects.

    Parameters
    ----------
    filename : str
        The filename of the file to access.

    cache : bool or 'update', optional
        Whether to cache the contents of remote URLs. If 'update', check
        the remote URL for a new version but store the result in the
        cache.

    Returns
    -------
    regions : `regions.Regions`
        A `Regions` object containing a list of `~regions.Region`
        objects.
    """
    with get_readable_fileobj(filename, cache=cache) as fh:
        region_string = fh.read()
        return _parse_ds9(region_string)


@RegionsRegistry.register(Regions, 'parse', 'ds9')
def _parse_ds9(region_str):
    """
    Parse a DS9 region string.

    Parameters
    ----------
    region_str : str
        The string contents of a DS9 region file.

    Returns
    -------
    regions : `regions.Regions`
        A `Regions` object containing a list of `~regions.Region`
        objects.
    """
    # first parse the input string to generate the raw region data
    region_data = _parse_raw_data(region_str)

    # now parse the raw region data into region object(s)
    regions = []
    for region_data_ in region_data:
        region = _make_region(region_data_)
        if region is not None:  # skip region if error during parsing
            regions.extend(region)
    return Regions(regions)


@dataclass
class _RegionData:
    """
    Class to hold data used to initialize a Region object.

    Data for multi-annulus regions is stored in a single object.
    """

    frame: str
    region_type: str
    shape: str
    shape_params: str
    raw_meta: dict
    region_str: str


def _split_lines(region_str):
    """
    Split a region string on newlines and all semicolons, except those
    between {} braces ({} contains ds9 text strings).

    Parameters
    ----------
    region_str : str
        The string contents of a DS9 region file.

    Returns
    -------
    lines : list of str
        A list of strings.
    """
    return [line_.strip() for line in region_str.split('\n')
            for line_ in _split_semicolon(line)]


def _parse_raw_data(region_str):
    """
    Parse the raw data to extract the region data.

    Parameters
    ----------
    region_str : str
        The string contents of a DS9 region file.

    Returns
    -------
    lines : list of `_RegionData`
        A list of `_RegionData` objects. Note that at this stage, region
        data for multi-annulus regions is stored in a single object.
    """
    global_meta = {}
    composite_meta = ''
    frame = None
    region_data = []

    supported_frames = ['image', 'icrs', 'fk5', 'j2000', 'fk4', 'b1950',
                        'galactic', 'ecliptic']
    unsupported_frames = ['linear', 'amplifier', 'detector', 'physical',
                          'tile']
    wcs_frames = ['wcs', 'wcs0'] + [f'wcs{letter}'
                                    for letter in string.ascii_lowercase]
    unsupported_frames += wcs_frames

    supported_shapes = ['circle', 'ellipse', 'box', 'annulus', 'polygon',
                        'line', 'point', 'text', 'composite']
    unsupported_shapes = ['vector', 'ruler', 'compass', 'projection',
                          'panda', 'epanda', 'bpanda']

    supported_frames_shapes = supported_frames + supported_shapes
    unsupported_frames_shapes = unsupported_frames + unsupported_shapes
    valid_frames_shapes = supported_frames_shapes + unsupported_frames_shapes

    regex_frame_or_shape = re.compile('^#? *([+-]?)([a-zA-Z0-9]+)')

    for line in _split_lines(region_str):  # split on semicolons & newlines
        # skip blank lines
        if not line:
            continue

        # skip comments
        if (line.startswith('#')
                and not line.startswith(('# text(', '# composite('))):
            continue

        original_line = line  # used to parse text and tag fields (keep case)
        line = line.lower()

        # ds9 region files can have multiple (including successive)
        # global lines
        if line.startswith('global'):
            global_meta.update(_parse_metadata(original_line[7:]))
            continue

        match = regex_frame_or_shape.search(line)
        if match is None:
            raise ValueError(f'Unable to parse line "{line}".')
        include_symbol = match.groups()[0]
        frame_or_shape = match.groups()[1]

        if frame_or_shape not in valid_frames_shapes:
            warnings.warn(f'"{frame_or_shape}" frame or shape is not a valid '
                          'frame or region shape; unable to parse line '
                          f'"{line}", skipping.', AstropyUserWarning)
            continue

        if frame_or_shape in unsupported_frames_shapes:
            warnings.warn(f'"{frame_or_shape}" frame or shape is not '
                          'supported by the regions package, skipping.',
                          AstropyUserWarning)
            if frame_or_shape in unsupported_frames:
                frame = None
            continue

        if frame_or_shape in supported_frames:
            # NOTE: frame value persists for subsequent regions until changed
            frame = frame_or_shape
            continue

        if frame_or_shape in supported_shapes:
            shape = frame_or_shape

            # a frame must be defined before any shape(s)
            if frame is None:
                warnings.warn('A coordinate frame was not found for region: '
                              f'"{line}", skipping.', AstropyUserWarning)
                continue

            if shape == 'composite':
                idx = line.find('||')
                if idx == -1:
                    raise ValueError('unable to parse line with composite '
                                     f'shape: "{line}"')
                # composite metadata applies to all regions within the
                # composite shape
                composite_meta = _parse_metadata(line[idx + 2:].strip())
                # remove "composite=1" since we split the composite
                composite_meta.pop('composite', None)

            # NOTE: include=1/0 in metadata overrides the leading
            #       "-/+" symbol; -: include=0;, + or '': include=1
            include = 0 if include_symbol == '-' else 1
            include_meta = {'include': include}

            params_str, meta_str = _parse_shape_line(shape, original_line,
                                                     match.span())

            region_meta = _parse_metadata(meta_str)
            raw_meta = _define_raw_metadata(global_meta, composite_meta,
                                            include_meta, region_meta)

            region_type = 'sky'
            if frame == 'image':
                region_type = 'pixel'

            # composite shape is used only to extract metadata
            if shape != 'composite':
                region_data.append(_RegionData(frame, region_type, shape,
                                               params_str, raw_meta, line))

            # reset composite metadata after the composite region ends
            if '||' not in line and composite_meta:
                composite_meta = {}

    return region_data


def _parse_shape_line(shape, line, span):
    """
    Parse a line of a DS9 file containing a shape.

    Parameters
    ----------
    shape : str
        The DS9 region shape (e.g., 'circle').

    line : str
        A line defining a DS9 shape.

    span : 2 tuple
        A tuple containing the (start, end) positions of the match
        defining the shape.

    Returns
    -------
    shape_params_str : str
        The region shape parameters as a string.

    meta_str : str
        The region metadata as a string.
    """
    full_line = line
    line = full_line[span[1]:]  # starts with the shape parameters

    # ds9 writes out text regions in this odd (undocumented) format
    if shape == 'text' and full_line.lower().startswith('# text'):
        idx = line.find(' ')
        if idx == -1:
            raise ValueError(f'unable to parse line "{line}"')
        meta_str = line[idx + 1:]
        shape_params_str = line[:idx]

    else:
        # split line into shape parameters and metadata
        parts = line.split('#', 1)  # color value can contain a #
        if len(parts) == 1:
            shape_params_str = line
            meta_str = ''  # no metadata found
        else:
            shape_params_str, meta_str = parts

    # strip trailing space and | chars
    shape_params_str = shape_params_str.strip(' |')
    shape_params_str = re.sub('[()]', '', shape_params_str).lower()
    meta_str = meta_str.strip()

    return shape_params_str, meta_str


def _parse_metadata(metadata_str):
    """
    Parse metadata for a single DS9 region.

    Parameters
    ----------
    metadata_str : str
        The region metadata (e.g., everything after the "#").

    Returns
    -------
    metadata : dict
        The region metadata as a dictionary.
    """
    # {.*?}    # all chars in curly braces
    # \'.*?\'  # all chars in single quotes
    # \".*?\"  # all chars in double quotes
    # [-?\d+\.?\d*\s]+\s?  # ([-/+]floats [whitespace]) incl. repeats
    #                        (e.g., dashlist=8 3, width=3, textangle=18.35)
    # [^=\s]+\s*[-?\d+\.?\d*]*   # (all chars (e.g., point=diamond) or
    #                              (all chars [whitespace] digits)
    #                                 (e.g., point=diamond 42)
    metadata_regex = re.compile(r'([a-zA-Z]+)\s*=\s*({.*?}|\'.*?\'|\".*?\"|'
                                r'[-?\d+\.?\d*\s]+\s?|'
                                r'[^=\s]+\s*[-?\d+\.?\d*]*)')

    metadata = {}
    for key, val in metadata_regex.findall(metadata_str):
        key = key.lower()
        val = val.strip().strip("'").strip('"').lstrip('{').rstrip('}')
        if key not in metadata:
            if key == 'tag':
                val = [val]  # tag value is always a list
            metadata[key] = val
        elif key == 'tag':
            metadata[key].append(val)
        else:
            warnings.warn(f'Found duplicate metadata for "{key}", '
                          'skipping', AstropyUserWarning)
    return metadata


def _define_raw_metadata(global_meta, composite_meta, include_meta,
                         region_meta):
    """
    Define the raw metadata dictionary for the region.

    The raw metadata is not modified, except that invalid metadata is
    ignored.

    Parameters
    ----------
    global_meta : dict
        The global metadata.

    composite_meta : dict
        The metadata for a composite region.

    include_meta : dict
        The include/exclude metadata.

    region_meta : dict
        The region metadata.

    Returns
    -------
    meta : tuple of dict
        The (valid) raw metadata extracted from the region file.
    """
    all_meta = global_meta.copy()
    all_meta.update(composite_meta)
    all_meta.update(include_meta)
    # region_meta must come after include_meta because include=1/0 in
    # metadata overrides the leading "-/+" include symbol
    all_meta.update(region_meta)

    # valid DS9 point symbols
    valid_points = ('circle', 'box', 'diamond', 'cross', 'x', 'arrow',
                    'boxcircle')

    # valid DS9 line values
    valid_lines = ('0 0', '0 1', '1 0', '1 1')

    # binary keys must have values of 0 or 1
    # NOTE: textrotate is not documented in the DS9 region file spec
    binary_keys = ('dash', 'select', 'highlite', 'fixed', 'edit', 'move',
                   'rotate', 'delete', 'include', 'source', 'background',
                   'fill', 'vector', 'textrotate')

    metadata = {}
    for key, value in all_meta.items():
        try:
            value = float(value)
            if value.is_integer():
                value = int(value)
        except (ValueError, TypeError):
            pass

        is_invalid = False
        # point value can either be ["symbol int"] or ["symbol"]
        if key == 'point':
            val = value.split()
            if val[0] not in valid_points:
                is_invalid = True
            if len(val) == 2 and not float(val[1]).is_integer():
                is_invalid = True

        if key == 'line' and value not in valid_lines:
            is_invalid = True

        if key in binary_keys and value not in (0, 1):
            is_invalid = True

        if is_invalid:
            warnings.warn(f'DS9 "{key}={value}" is invalid and will be '
                          'ignored', AstropyUserWarning)
            continue

        metadata[key] = value

    return metadata


def _parse_pixel_coord(param_str):
    invalid_chars = ('d', 'r', 'p', ':', 'h', 'm', 's')
    for char in invalid_chars:
        if char in param_str:
            raise DS9ParserError('Cannot parse pixel region position '
                                 'coordinates')

    if param_str[-1] == 'i':
        param_str = param_str[:-1]

    # DS9 uses 1-indexed pixels
    return float(param_str) - 1


def _parse_sky_coord(param_str, frame, index):
    invalid_chars = ('i', 'p')
    for char in invalid_chars:
        if char in param_str:
            raise DS9ParserError('Cannot parse sky region position '
                                 'coordinates')

    if param_str[-1] == 'r':
        return Angle(param_str[:-1], unit=u.radian)

    elif param_str[-1] == 'd':
        return Angle(param_str[:-1], unit=u.degree)

    elif 'd' in param_str or 'h' in param_str:
        return Angle(param_str)

    elif ':' in param_str:
        if index % 2 == 0 and frame not in ('galactic', 'ecliptic'):
            return Angle(param_str, u.hourangle)
        else:
            return Angle(param_str, u.degree)

    else:  # unit not specified
        return Angle(float(param_str), unit=u.degree)


def _parse_coord(region_type, param_str, frame, index):
    if region_type == 'pixel':
        return _parse_pixel_coord(param_str)
    else:
        return _parse_sky_coord(param_str, frame, index)


def _parse_angle(param_str):
    invalid_chars = ('p', 'i')
    for char in invalid_chars:
        if char in param_str:
            raise DS9ParserError('Cannot parse sky region angle parameter')

    unit_mapping = {'"': u.arcsec,
                    "'": u.arcmin,
                    'd': u.deg,
                    'r': u.rad}
    if param_str[-1] not in string.digits:
        unit = unit_mapping[param_str[-1]]
        return u.Quantity(float(param_str[:-1]), unit=unit)
    else:
        return u.Quantity(float(param_str), unit=u.degree)


def _parse_size(region_type, param_str):
    if region_type == 'pixel':
        invalid_chars = ('"', "'", 'd', 'r', 'p')
        for char in invalid_chars:
            if char in param_str:
                raise DS9ParserError('Cannot parse pixel region size '
                                     'parameters - must not be in angular '
                                     'units')

        if param_str[-1] == 'i':
            param_str = param_str[:-1]

        return float(param_str)

    else:
        # size in angular units
        return _parse_angle(param_str)


def _parse_shape_params(region_data):
    """
    Parse the shape parameters for a region line.

    Parameters
    ----------
    region_data : `_RegionData` instance
        A `_RegionData` instance containing the data for a region line.
        Data for multi-annulus regions is stored in a single object.

    Returns
    -------
    shape_params : list of list(s)
        The shape parameters for each region(s). ``shape_params`` is
        a list of lists of shape parameters. A separate shape list is
        returned for multiple annulus/box/ellipse regions. Otherwise,
        the output list contains only one list.
    """
    region_type = region_data.region_type
    shape = region_data.shape
    frame = region_data.frame
    params = [val for val in re.split(r'\s|\,', region_data.shape_params)
              if val]  # split values on space or comma

    nparams = len(params)
    n_annulus = 0

    if shape in ('ellipse', 'box') and nparams > 5:
        if nparams % 2 != 1:
            raise ValueError(f'incorrect number of parameters ({nparams}) '
                             f'for shape "{shape}"')
        n_annulus = ((nparams - 3) // 2) - 1

    if shape in ('annulus',):
        n_annulus = nparams - 3

    if shape in ('ellipse', 'box', 'annulus'):
        # reset the cycle iterators
        shape_template = make_region_template()
    else:
        shape_template = ds9_params_template[shape]

    shape_params = []
    # TODO: check zip strict=True
    for idx, (param_type, value) in enumerate(zip(shape_template, params)):
        if shape in ('ellipse', 'box') and idx == nparams - 1:
            param_type = 'angle'  # last parameter is always an angle

        if param_type == 'coord':
            param = _parse_coord(region_type, value, frame, idx)
        elif param_type in ('length',):
            param = _parse_size(region_type, value)
            if shape == 'ellipse':
                param *= 2.0  # ds9 uses semi-axis lengths
        elif param_type in ('angle',):
            param = _parse_angle(value)
        else:
            raise ValueError('cannot parse shape parameters')

        shape_params.append(param)

    if n_annulus > 1:
        tmp_params = []
        for i in range(n_annulus):
            if shape in ('ellipse', 'box'):
                idx = (i + 1) * 2
                params = [shape_params[0], shape_params[1], shape_params[idx],
                          shape_params[idx + 1], shape_params[idx + 2],
                          shape_params[idx + 3], shape_params[-1]]
            else:
                idx = i + 2
                params = [shape_params[0], shape_params[1], shape_params[idx],
                          shape_params[idx + 1]]

            tmp_params.append(params)
        shape_params = tmp_params
    else:
        shape_params = [shape_params]

    if n_annulus > 0:
        if shape == 'ellipse':
            shape = 'ellipse_annulus'
        elif shape == 'box':
            shape = 'rectangle_annulus'

    return shape, shape_params


def _define_coords(region_type, params, frame=None):
    if region_type == 'pixel':
        coords = PixCoord(*params)
    else:
        coords = SkyCoord(*params, frame=frame)
    return coords


def _define_region_params(region_type, shape, shape_params, frame=None):
    if frame is not None:
        frame = ds9_frame_map[frame]

    if shape == 'polygon':
        coord_params = (shape_params[0::2], shape_params[1::2])
        params = [_define_coords(region_type, coord_params, frame=frame)]

    elif shape == 'line':
        params = [_define_coords(region_type, shape_params[0:2], frame=frame),
                  _define_coords(region_type, shape_params[2:4], frame=frame)]

    elif shape in ('ellipse_annulus', 'rectangle_annulus'):
        size_params = shape_params[2:-1]
        tmp = [size_params[0::2], size_params[1::2]]
        tmp_flat = [item for sublist in tmp for item in sublist]
        params = [_define_coords(region_type, shape_params[0:2], frame=frame),
                  *tmp_flat, shape_params[-1]]

    else:
        params = ([_define_coords(region_type, shape_params[0:2],
                                  frame=frame)] + shape_params[2:])

    return params


def _make_region(region_data):
    """
    Make a region object from the region data.

    Parameters
    ----------
    region_data : `_RegionData` instance
        A `_RegionData` instance.
    """
    try:
        # NOTE: returned shape can be different from region_data.shape
        shape, shape_params_list = _parse_shape_params(region_data)
    except DS9ParserError as err:
        # raise a warning and skip the region
        msg = f'{str(err)}: {region_data.region_str}'
        warnings.warn(msg, AstropyUserWarning)
        return None

    # define the parameters to initialize a Region
    # NOTE: region_params can be longer than region_data for
    # multi-annulus regions
    region_type = region_data.region_type
    region_params = []
    for shape_params in shape_params_list:
        region_params.extend([_define_region_params(region_type, shape,
                                                    shape_params,
                                                    region_data.frame)])

    # separate the metadata and visual metadata and then translate the
    # visual metadata to valid mpl kwargs for the particular region
    meta, visual = _split_raw_metadata(region_data.raw_meta)
    visual = _translate_ds9_to_visual(shape, visual)

    regions = []
    for shape_params in region_params:
        # for Text region, we need to add meta['text'] to params and
        # remove it from meta; set to '' if the text meta value was not
        # specified
        if shape == 'text':
            shape_params.append(region_data.raw_meta.get('text', ''))
            meta.pop('text', None)

        region = ds9_shape_to_region[region_type][shape](*shape_params)

        region.meta = RegionMeta(meta)
        region.visual = RegionVisual(visual)
        region._raw_meta = region_data.raw_meta

        regions.append(region)

    return regions


def _find_text_delim_idx(region_str):
    """
    Find the indices of the DS9 text field delimiters ({}, '', or "") in
    a string.
    """
    pattern = re.compile(r'(text\s*=\s*[{\'"])')
    idx0 = []
    delim = []
    start_idx = []
    for match in pattern.finditer(region_str):
        idx0.append(match.start())
        end_delim = match.group()[-1]
        if end_delim == '{':
            end_delim = '}'
        delim.append(end_delim)
        start_idx.append(match.span()[1])

    idx1 = []
    for sidx, char in zip(start_idx, delim, strict=True):
        idx1.append(region_str.find(char, sidx))

    return idx0, idx1


def _split_semicolon(region_str):
    r"""
    Split a DS9 region string on semicolons.

    The line is not split on semicolons found in a text field.

    This turned out to be a very tricky problem (attempts with regex
    failed):

      * text fields are delimited by {}, '', or ""
      * the text delimiters do not have to be consistent within a file
      * region strings can have unpaired ' (arcmin) or " (arcsec)
      * region strings can have "#" in a color value (cannot split on #)
      * the text field can contain "text={str}", "text='str'",
        'test="str"' etc., e.g., text={text="hello"}
      * escaped delimiter characters used by the text field
        are not allowed within the text field, e.g., "text={my
        field\{test\}}" and "text='my field \'test\''" are invalid.
        However, "text={my field, 'test'}" is valid

    This code finds the text field delimiters and then finds the indices
    of the opening and closing delimiters. Text fields that contain
    "text={str}", etc. are included (as a smaller range between the
    larger text field range), but this is fine because all we need are
    index ranges where to exclude semicolons (for splitting). Semicolons
    found at indices between the open/close delimiter indices are
    excluded from splitting.
    """
    idx0, idx1 = _find_text_delim_idx(region_str)

    semi_idx = [pos for pos, char in enumerate(region_str) if char == ';']
    fidx = []
    for i in semi_idx:
        for i0, i1 in zip(idx0, idx1, strict=True):
            if i0 <= i <= i1:
                break
        else:
            fidx.append(i + 1)
    fidx.insert(0, 0)

    return [region_str[i:j].rstrip(';')
            for i, j in zip(fidx, fidx[1:] + [None], strict=True)]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
#
# Documentation build configuration file.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# file.
#
# All configuration values have a default. Some values are defined in
# the global Astropy configuration which is loaded here before anything
# else. See astropy.sphinx.conf for which values are set there.

import os
import sys
import tomllib
from datetime import UTC, datetime
from importlib import metadata
from pathlib import Path

try:
    from sphinx_astropy.conf.v1 import *  # noqa: F403
    from sphinx_astropy.conf.v1 import rst_epilog
except ImportError:
    print('ERROR: the documentation requires the sphinx-astropy package to '
          'be installed')
    sys.exit(1)

# Get configuration information from pyproject.toml
with (Path(__file__).parents[1] / 'pyproject.toml').open('rb') as fh:
    project_meta = tomllib.load(fh)['project']

# -- General configuration ----------------------------------------------------
# By default, highlight as Python 3.
highlight_language = 'python3'

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '3.0'

# Extend astropy intersphinx_mapping with packages we use here
intersphinx_mapping.update(  # noqa: F405
    {'photutils': ('https://photutils.readthedocs.io/en/stable/', None),
     # 'shapely': ('https://shapely.readthedocs.io/en/stable/', None),
     })

# Exclude astropy intersphinx_mapping for unused packages
del intersphinx_mapping['scipy']  # noqa: F405
del intersphinx_mapping['h5py']  # noqa: F405

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# exclude_patterns.append('_templates')

plot_formats = ['png', 'hires.png', 'pdf', 'svg']

# This is added to the end of RST files - a good place to put
# substitutions to be used globally.
with open('common_links.txt') as fh:
    rst_epilog += fh.read()

# -- Project information ------------------------------------------------------
project = project_meta['name']
author = project_meta['authors'][0]['name']
project_copyright = f'2015-{datetime.now(tz=UTC).year}, {author}'
github_project = 'astropy/regions'

# The version info for the project you're documenting, acts as
# replacement for |version| and |release|, also used in various other
# places throughout the built documents.
release = metadata.version(project)
# The short X.Y version.
version = '.'.join(release.split('.')[:2])
dev = 'dev' in release

# -- Options for HTML output --------------------------------------------------
# The global astropy configuration uses a custom theme,
# 'bootstrap-astropy', which is installed along with astropy. A
# different theme can be used or the options for this theme can be
# modified by overriding some of the variables set in the global
# configuration. The variables set in the global configuration are
# listed below, commented out.

# Add any paths that contain custom themes here, relative to this
# directory.
# html_theme_path = []

# The theme to use for HTML and HTML Help pages. See the documentation
# for a list of builtin themes. To override the custom theme, set this
# to the name of a builtin theme or the name of a custom theme in
# html_theme_path.
# html_theme = None

# Customized theme options
html_theme_options = {
    'logotext1': 'regions',  # white,  semi-bold
    'logotext2': '',  # orange, light
    'logotext3': ':docs'   # white,  light
}

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# The name of an image file (relative to this directory) to place at the
# top of the sidebar.
# html_logo = ''

# The name of an image file (within the static path) to use as favicon
# of the docs. This file should be a Windows icon file (.ico) being
# 16x16 or 32x32 pixels large.
# html_favicon = ''

# A "Last built" timestamp is inserted at every page bottom, using the
# given strftime format. Set to '' to omit this timestamp.
# html_last_updated_fmt = '%d %b %Y'

# The name for this set of Sphinx documents. If None, it defaults to
# "<project> v<release>".
html_title = f'{project} {release}'

# Output file base name for HTML help builder.
htmlhelp_basename = project + 'doc'

# Static files to copy after template files
# html_static_path = ['_static']
# html_style = 'regions.css'

# Set canonical URL from the Read the Docs Domain
html_baseurl = os.environ.get('READTHEDOCS_CANONICAL_URL', '')

# A dictionary of values to pass into the template engine's context for
# all pages.
html_context = {
    'default_mode': 'light',
    'to_be_indexed': ['stable', 'latest'],
    'is_development': dev,
    'github_user': 'astropy',
    'github_repo': 'regions',
    'github_version': 'main',
    'doc_path': 'docs',
    # Tell Jinja2 templates the build is running on Read the Docs
    'READTHEDOCS': os.environ.get('READTHEDOCS', '') == 'True',
}

# -- Options for LaTeX output -------------------------------------------------
# Grouping the document tree into LaTeX files. List of tuples (source
# start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [('index', project + '.tex', project + ' Documentation',
                    author, 'manual')]
# latex_logo = '_static/regions_banner.pdf'

# -- Options for manual page output -------------------------------------------
# One entry per manual page. List of tuples (source start file, name,
# description, authors, manual section).
man_pages = [('index', project.lower(), project + ' Documentation',
              [author], 1)]

# -- Resolving issue number to links in changelog -----------------------------
github_issues_url = f'https://github.com/{github_project}/issues/'

# -- Turn on nitpicky mode for sphinx (to warn about references not found) ----
nitpicky = True
nitpick_ignore = []

# Some warnings are impossible to suppress, and you can list specific
# references that should be ignored in a nitpick-exceptions file which
# should be inside the docs/ directory. The format of the file should be:
#
# <type> <class>
#
# for example:
#
# py:class astropy.io.votable.tree.Element
# py:class astropy.io.votable.tree.SimpleElement
# py:class astropy.io.votable.tree.SimpleElementWithContent
#
# Uncomment the following lines to enable the exceptions:
nitpick_filename = 'nitpick-exceptions.txt'
if os.path.isfile(nitpick_filename):
    with open(nitpick_filename) as fh:
        for line in fh:
            if line.strip() == '' or line.startswith('#'):
                continue
            dtype, target = line.split(None, 1)
            target = target.strip()
            nitpick_ignore.append((dtype, target))

# -- Options for linkcheck output ---------------------------------------------
linkcheck_retry = 5
linkcheck_ignore = ['http://data.astropy.org',
                    r'https://github\.com/astropy/regions/(?:issues|pull)/\d+']
linkcheck_timeout = 180

# -- Matplotlib plot defaults -------------------------------------------------
plot_rcparams = {'savefig.bbox': 'tight',
                 'savefig.facecolor': 'none',
                 'axes.formatter.useoffset': False,
                 'xtick.labelsize': 9,
                 'ytick.labelsize': 9,
                 'axes.labelsize': 11,
                 'figure.titlesize': 11,
                 'figure.subplot.wspace': 0.23,
                 'figure.subplot.hspace': 0.23}
plot_apply_rcparams = True



# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure. It needs to live inside the package in order for it to
# get picked up when running the tests inside an interpreter using
# packagename.test

import numpy as np
from astropy.utils import minversion

try:
    from pytest_astropy_header.display import (PYTEST_HEADER_MODULES,
                                               TESTED_VERSIONS)
    ASTROPY_HEADER = True
except ImportError:
    ASTROPY_HEADER = False


# do not remove until we drop support for NumPy < 2.0
if minversion(np, '2.0.0.dev0+git20230726'):
    np.set_printoptions(legacy='1.25')


def pytest_configure(config):
    if ASTROPY_HEADER:
        config.option.astropy_header = True

        # Customize the following lines to add/remove entries from the
        # list of packages for which version numbers are displayed when
        # running the tests.
        PYTEST_HEADER_MODULES.clear()
        deps = ['NumPy', 'Matplotlib', 'Astropy', 'Shapely']
        for dep in deps:
            PYTEST_HEADER_MODULES[dep] = dep.lower()

        from regions import __version__
        TESTED_VERSIONS['regions'] = __version__



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
The module provides several custom descriptor classes for attribute
validation of region classes.
"""

import abc

import numpy as np
from astropy.coordinates import SkyCoord
from astropy.units import Quantity

from regions.core.metadata import RegionMeta, RegionVisual
from regions.core.pixcoord import PixCoord

__all__ = []


class RegionAttribute(abc.ABC):
    """
    Base descriptor class for region attribute validation.

    Parameters
    ----------
    doc : str, optional
        The documentation string for the attribute.
    """

    def __init__(self, doc=''):
        self.__doc__ = doc

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self  # pragma: no cover
        return instance.__dict__[self.name]

    def __set__(self, instance, value):
        self._validate(value)
        instance.__dict__[self.name] = value

    def __delete__(self, instance):
        raise AttributeError(f'cannot delete {self.name!r}')

    @abc.abstractmethod
    def _validate(self, value):
        """
        Validate the attribute value.

        An exception is raised if the value is invalid.
        """
        raise NotImplementedError  # pragma: no cover


class ScalarPixCoord(RegionAttribute):
    """
    Descriptor class to check that value is a scalar
    `~regions.PixCoord`.
    """

    def _validate(self, value):
        if not (isinstance(value, PixCoord) and value.isscalar):
            raise ValueError(f'{self.name!r} must be a scalar PixCoord')


class OneDPixCoord(RegionAttribute):
    """
    Descriptor class to check that value is a 1D `~regions.PixCoord`.
    """

    def _validate(self, value):
        if not (isinstance(value, PixCoord) and not value.isscalar
                and value.x.ndim == 1):
            raise ValueError(f'{self.name!r} must be a 1D PixCoord')


class PositiveScalar(RegionAttribute):
    """
    Descriptor class to check that value is a strictly positive (> 0)
    scalar float/int (not `~astropy.units.Quantity`).
    """

    def _validate(self, value):
        if isinstance(value, Quantity):
            raise ValueError(f'{self.name!r} must be a scalar integer or '
                             'float')

        if not np.isscalar(value) or value <= 0:
            raise ValueError(f'{self.name!r} must be a strictly positive '
                             'scalar')


class ScalarSkyCoord(RegionAttribute):
    """
    Descriptor class to check that value is a scalar
    `~astropy.coordinates.SkyCoord`.
    """

    def _validate(self, value):
        if not (isinstance(value, SkyCoord) and value.isscalar):
            raise ValueError(f'{self.name!r} must be a scalar SkyCoord')


class OneDSkyCoord(RegionAttribute):
    """
    Descriptor class to check that value is a 1D
    `~astropy.coordinates.SkyCoord`.
    """

    def _validate(self, value):
        if not (isinstance(value, SkyCoord) and value.ndim == 1):
            raise ValueError(f'{self.name!r} must be a 1D SkyCoord')


class ScalarAngle(RegionAttribute):
    """
    Descriptor class to check that value is a scalar angle, either an
    `~astropy.coordinates.Angle` or `~astropy.units.Quantity` with
    angular units.
    """

    def _validate(self, value):
        if isinstance(value, Quantity):
            if not value.isscalar:
                raise ValueError(f'{self.name!r} must be a scalar')

            if not value.unit.physical_type == 'angle':
                raise ValueError(f'{self.name!r} must have angular units')
        else:
            raise ValueError(f'{self.name!r} must be a scalar angle')


class PositiveScalarAngle(RegionAttribute):
    """
    Descriptor class to check that value is a strictly positive scalar
    angle, either an `~astropy.coordinates.Angle` or
    `~astropy.units.Quantity` with angular units.
    """

    def _validate(self, value):
        if isinstance(value, Quantity):
            if not value.isscalar:
                raise ValueError(f'{self.name!r} must be a scalar')

            if not value.unit.physical_type == 'angle':
                raise ValueError(f'{self.name!r} must have angular units')

            if not value > 0:
                raise ValueError(f'{self.name!r} must be strictly positive')
        else:
            raise ValueError(f'{self.name!r} must be a strictly positive '
                             'scalar angle')


class RegionType(RegionAttribute):
    """
    Descriptor class to check the region type of value.

    Parameters
    ----------
    name : str
        The name of the attribute.

    regionclass : `~regions.Region`
        The region class to check.
    """

    def __init__(self, name, regionclass):
        super().__init__(name)
        self.regionclass = regionclass

    def _validate(self, value):
        if not isinstance(value, self.regionclass):
            raise ValueError(f'{self.name!r} must be a '
                             f'{self.regionclass.__name__} object')


class RegionMetaDescr(RegionAttribute):
    """
    Descriptor class for the region meta dictionary.

    If input as a pure `dict`, it will be converted to a `RegionMeta`
    object.
    """

    def __set__(self, instance, value):
        # RegionMeta subclasses dict
        if isinstance(value, dict) and not isinstance(value, RegionMeta):
            value = RegionMeta(value)
        super().__set__(instance, value)

    def _validate(self, value):
        if not isinstance(value, RegionMeta):
            raise ValueError(f'{self.name!r} must be a dict or RegionMeta '
                             'object')


class RegionVisualDescr(RegionAttribute):
    """
    Descriptor class for the region visual dictionary.

    If input as a pure `dict`, it will be converted to a `RegionVisual`
    object.
    """

    def __set__(self, instance, value):
        # RegionVisual subclasses dict
        if isinstance(value, dict) and not isinstance(value, RegionVisual):
            value = RegionVisual(value)
        super().__set__(instance, value)

    def _validate(self, value):
        if not isinstance(value, RegionVisual):
            raise ValueError(f'{self.name!r} must be a dict or RegionVisual '
                             'object')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines polygon regions in both pixel and sky coordinates.
"""

import astropy.units as u
import numpy as np

from regions._geometry import polygonal_overlap_grid
from regions._geometry.pnpoly import points_in_polygon
from regions.core.attributes import (OneDPixCoord, OneDSkyCoord,
                                     PositiveScalar, RegionMetaDescr,
                                     RegionVisualDescr, ScalarAngle,
                                     ScalarPixCoord)
from regions.core.bounding_box import RegionBoundingBox
from regions.core.core import PixelRegion, SkyRegion
from regions.core.mask import RegionMask
from regions.core.metadata import RegionMeta, RegionVisual
from regions.core.pixcoord import PixCoord

__all__ = ['PolygonPixelRegion', 'RegularPolygonPixelRegion',
           'PolygonSkyRegion']


class PolygonPixelRegion(PixelRegion):
    """
    A polygon in pixel coordinates.

    Parameters
    ----------
    vertices : `~regions.PixCoord`
        The vertices of the polygon.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.
    origin : `~regions.PixCoord`, optional
        The origin for polynomial vertices. Using this keyword allows
        ``vertices`` to be specified relative to an origin pixel
        coordinate.

    Examples
    --------
    .. plot::
        :include-source:

        from regions import PixCoord, PolygonPixelRegion
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1)

        x, y = [45, 45, 55, 60], [75, 70, 65, 75]
        vertices = PixCoord(x=x, y=y)
        reg = PolygonPixelRegion(vertices=vertices)
        patch = reg.plot(ax=ax, facecolor='none', edgecolor='red', lw=2,
                         label='Polygon')

        ax.legend(handles=(patch,), loc='upper center')
        ax.set_xlim(30, 80)
        ax.set_ylim(50, 80)
        ax.set_aspect('equal')
    """

    _params = ('vertices',)
    _mpl_artist = 'Patch'
    vertices = OneDPixCoord('The vertices of the polygon as a |PixCoord| '
                            'array.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, vertices, meta=None, visual=None, origin=None):
        self._vertices = vertices
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()
        if origin is None:
            origin = PixCoord(0, 0)
        self.origin = origin
        self.vertices = vertices + origin

    @property
    def area(self):
        # See https://stackoverflow.com/questions/24467972

        # Use offsets to improve numerical precision
        x_ = self.vertices.x - self.vertices.x.mean()
        y_ = self.vertices.y - self.vertices.y.mean()

        # Shoelace formula, for our case where the start vertex
        # isn't duplicated at the end, written to avoid an array copy
        area_main = np.dot(x_[:-1], y_[1:]) - np.dot(y_[:-1], x_[1:])
        area_last = x_[-1] * y_[0] - y_[-1] * x_[0]
        return 0.5 * np.abs(area_main + area_last)

    def contains(self, pixcoord):
        pixcoord = PixCoord._validate(pixcoord, 'pixcoord')
        x = np.atleast_1d(np.asarray(pixcoord.x, dtype=float))
        y = np.atleast_1d(np.asarray(pixcoord.y, dtype=float))
        vx = np.asarray(self.vertices.x, dtype=float)
        vy = np.asarray(self.vertices.y, dtype=float)

        shape = x.shape
        mask = points_in_polygon(x.flatten(), y.flatten(), vx, vy).astype(bool)
        in_poly = mask.reshape(shape)
        if self.meta.get('include', True):
            return in_poly
        else:
            return np.logical_not(in_poly)

    def to_sky(self, wcs):
        vertices_sky = wcs.pixel_to_world(self.vertices.x, self.vertices.y)
        return PolygonSkyRegion(vertices=vertices_sky, meta=self.meta.copy(),
                                visual=self.visual.copy())

    @property
    def bounding_box(self):
        xmin = self.vertices.x.min()
        xmax = self.vertices.x.max()
        ymin = self.vertices.y.min()
        ymax = self.vertices.y.max()
        return RegionBoundingBox.from_float(xmin, xmax, ymin, ymax)

    def to_mask(self, mode='center', subpixels=5):
        self._validate_mode(mode, subpixels)

        if mode == 'center':
            mode = 'subpixels'
            subpixels = 1

        use_exact = 0 if mode == 'subpixels' else 1

        # Find bounding box and mask size
        bbox = self.bounding_box
        ny, nx = bbox.shape

        # Find position of pixel edges and recenter so that circle is at
        # origin
        xmin = float(bbox.ixmin) - 0.5
        xmax = float(bbox.ixmax) - 0.5
        ymin = float(bbox.iymin) - 0.5
        ymax = float(bbox.iymax) - 0.5

        vx = np.asarray(self.vertices.x, dtype=float)
        vy = np.asarray(self.vertices.y, dtype=float)

        fraction = polygonal_overlap_grid(xmin, xmax, ymin, ymax, nx, ny,
                                          vx, vy, use_exact, subpixels)

        return RegionMask(fraction, bbox=bbox)

    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Return a matplotlib patch object for this region
        (`matplotlib.patches.Polygon`).

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        **kwargs : dict
            Any keyword arguments accepted by
            `~matplotlib.patches.Polygon`. These keywords will override
            any visual meta attributes of this region.

        Returns
        -------
        artist : `~matplotlib.patches.Polygon`
            A matplotlib polygon patch.
        """
        from matplotlib.patches import Polygon

        xy = np.vstack([self.vertices.x - origin[0],
                        self.vertices.y - origin[1]]).transpose()

        mpl_kwargs = self.visual.define_mpl_kwargs(self._mpl_artist)
        mpl_kwargs.update(kwargs)

        return Polygon(xy=xy, **mpl_kwargs)

    def rotate(self, center, angle):
        """
        Rotate the region.

        Positive ``angle`` corresponds to counter-clockwise rotation.

        Parameters
        ----------
        center : `~regions.PixCoord`
            The rotation center point.
        angle : `~astropy.coordinates.Angle`
            The rotation angle.

        Returns
        -------
        region : `PolygonPixelRegion`
            The rotated region (which is an independent copy).
        """
        vertices = self.vertices.rotate(center, angle)
        return self.copy(vertices=vertices)


class RegularPolygonPixelRegion(PolygonPixelRegion):
    """
    A regular polygon in pixel coordinates.

    .. note::

        This class will be serialized as a generic polygon
        region, thus when read back in it will produce a
        `~regions.PolygonPixelRegion` object instead of a
        `~regions.RegularPolygonPixelRegion` object.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the polygon.
    nvertices : `~regions.PixCoord`
        The number of polygon vertices (or sides).
    radius : float
        The distance from the center to any vertex. This is also known
        as the circumradius.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the polygon, measured anti-clockwise. If
        set to zero (the default), the polygon will point "up" following
        the `matplotlib.patches.RegularPolygon` convention.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.

    Attributes
    ----------
    side_length : float
        The side length.
    inradius : float
        The radius of the largest circle contained entirely within the
        polygon. This value is identical to the length of a line segment
        from the polygon center to the midpoint of one of its sides
        (known as as the apothem).
    perimeter : float
        The polygon perimeter.
    interior_angle : float
        The polygon interior angle, which is the angle at each vertex on
        the inside of the polygon.
    exterior_angle : float
        The polygon exterior angle, which is an angle at each vertex on
        the outside of the polygon.

    Examples
    --------
    .. plot::
        :include-source:

        import astropy.units as u
        import matplotlib.pyplot as plt
        from regions import PixCoord, RegularPolygonPixelRegion

        fig, ax = plt.subplots(1, 1)
        center = PixCoord(x=50, y=50)
        reg1 = RegularPolygonPixelRegion(center, 6, 15)
        reg1.plot(edgecolor='red', lw=2)

        center = PixCoord(x=25, y=25)
        reg2 = RegularPolygonPixelRegion(center, 3, 15)
        reg2.plot(edgecolor='green', lw=2)

        center = PixCoord(x=25, y=75)
        reg3 = RegularPolygonPixelRegion(center, 3, 15, angle=25*u.deg)
        reg3.plot(edgecolor='orange', lw=2)

        center = PixCoord(x=75, y=75)
        reg4 = RegularPolygonPixelRegion(center, 8, 15)
        reg4.plot(edgecolor='blue', lw=2)

        center = PixCoord(x=75, y=25)
        reg5 = RegularPolygonPixelRegion(center, 5, 15)
        reg5.plot(edgecolor='magenta', lw=2)

        ax.set_xlim(0, 100)
        ax.set_ylim(0, 100)
        ax.set_aspect('equal')
    """

    _params = ('center', 'nvertices', 'radius', 'angle')
    center = ScalarPixCoord('The center pixel position as a |PixCoord|.')
    nvertices = PositiveScalar('The number of polygon vertices.')
    radius = PositiveScalar('The distance from the center to any vertex in '
                            'pixels as a float.')
    angle = ScalarAngle('The rotation angle measured anti-clockwise as a '
                        '|Quantity| angle.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, center, nvertices, radius, angle=0. * u.deg,
                 meta=None, visual=None):

        if nvertices < 3:
            raise ValueError('nvertices must be >= 3')
        self.center = center
        self.nvertices = nvertices
        self.radius = radius
        self.angle = angle

        super().__init__(self._calc_vertices(), meta=meta, visual=visual)

        self.side_length = 2. * self.radius * np.sin(np.pi / self.nvertices)
        self.inradius = self.radius * np.cos(np.pi / self.nvertices)
        self.perimeter = self.side_length * self.nvertices
        self.interior_angle = ((self.nvertices - 2) / self.nvertices
                               * 180) << u.deg
        self.exterior_angle = 360. / self.nvertices << u.deg

    def _calc_vertices(self):
        # uses the matplotlib convention that the polygon always points "up"
        theta = ((2. * np.pi / self.nvertices * np.arange(self.nvertices)
                 + (np.pi / 2)) << u.radian) + self.angle
        xvert = self.radius * np.cos(theta)
        yvert = self.radius * np.sin(theta)
        return self.center + PixCoord(xvert, yvert)

    def rotate(self, center, angle):
        """
        Rotate the region.

        Positive ``angle`` corresponds to counter-clockwise rotation.

        Parameters
        ----------
        center : `~regions.PixCoord`
            The rotation center point.
        angle : `~astropy.coordinates.Angle`
            The rotation angle.

        Returns
        -------
        region : `PolygonPixelRegion`
            The rotated region (which is an independent copy).
        """
        center = self.center.rotate(center, angle)
        angle = self.angle + angle
        return self.copy(center=center, angle=angle)

    def to_polygon(self):
        """
        Return a `PolygonPixelRegion` of this region.
        """
        return PolygonPixelRegion(self.vertices, origin=PixCoord(0, 0),
                                  meta=self.meta.copy(),
                                  visual=self.visual.copy())


class PolygonSkyRegion(SkyRegion):
    """
    A polygon defined using vertices in sky coordinates.

    Parameters
    ----------
    vertices : `~astropy.coordinates.SkyCoord`
        The vertices of the polygon.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.
    """

    _params = ('vertices',)
    vertices = OneDSkyCoord('The vertices of the polygon as a |SkyCoord| '
                            'array.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, vertices, meta=None, visual=None):
        self.vertices = vertices
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    def to_pixel(self, wcs):
        x, y = wcs.world_to_pixel(self.vertices)
        vertices_pix = PixCoord(x, y)
        return PolygonPixelRegion(vertices_pix, meta=self.meta.copy(),
                                  visual=self.visual.copy())



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.tests.helper import assert_quantity_allclose
from astropy.wcs import WCS
from numpy.testing import assert_allclose

from regions.core import PixCoord, Region


def make_simple_wcs(skycoord, resolution, size):
    """
    Create a simple WCS object based on a SkyCoord, resolution, and
    size.

    Parameters
    ----------
    skycoord : `~astropy.coordinates.SkyCoord`
        The sky coordinate for the center of the WCS.

    resolution : `~astropy.units.Quantity`
        The pixel resolution in degrees.

    size : int
        The size of the WCS in pixels (assumed square).

    Returns
    -------
    wcs : `~astropy.wcs.WCS`
        A WCS object with the specified parameters.
    """
    crpix = (size + 1) / 2
    cdelt = resolution.to(u.deg).value
    skycoord_icrs = skycoord.transform_to('icrs')
    ra = skycoord_icrs.ra.degree
    dec = skycoord_icrs.dec.degree

    wcs = WCS(naxis=2)
    wcs.wcs.crpix = [crpix, crpix]
    wcs.wcs.cdelt = np.array([-cdelt, cdelt])
    wcs.wcs.crval = [ra, dec]
    wcs.wcs.ctype = ['RA---TAN', 'DEC--TAN']

    return wcs


def assert_skycoord_allclose(skycoord1, skycoord2, **kwargs):
    """
    Test that two SkyCoord objects are nearly equal.

    Parameters
    ----------
    skycoord1 : `~astropy.coordinates.SkyCoord`
        First SkyCoord object to compare.

    skycoord2 : `~astropy.coordinates.SkyCoord`
        Second SkyCoord object to compare.

    **kwargs : dict
        Additional keyword arguments passed to
        `astropy.tests.assert_quantity_allclose`.
    """
    for attr in (skycoord1._extra_frameattr_names
                 | skycoord2._extra_frameattr_names):
        if not skycoord1.frame._frameattr_equiv(getattr(skycoord1, attr),
                                                getattr(skycoord2, attr)):
            raise ValueError('Extra frame attributes are not equivalent')

    assert skycoord1.is_equivalent_frame(skycoord2)

    for attr in (set(skycoord1._sky_coord_frame._data.components)
                 | set(skycoord2._sky_coord_frame._data.components)):
        sdata1 = skycoord1._sky_coord_frame._data
        sdata2 = skycoord2._sky_coord_frame._data
        assert_quantity_allclose(getattr(sdata1, attr), getattr(sdata2, attr),
                                 **kwargs)


def assert_region_allclose(region1, region2, **kwargs):
    """
    Test that two Region objects have parameters which are nearly equal.

    Meta and Visual properties are matched identically.

    Parameters
    ----------
    region1 : `~regions.core.Region`
        First Region object to compare.

    region2 : `~regions.core.Region`
        Second Region object to compare.

    **kwargs : dict
        Additional keyword arguments passed to
        `astropy.tests.assert_quantity_allclose` or similar functions
        for comparing specific parameter types.
    """
    if not (isinstance(region1, Region) and isinstance(region2, Region)):
        raise TypeError('Both inputs must be Region instances')

    # check that both have identical parameters
    region1_params = list(region1._params)
    region2_params = list(region2._params)
    if region1_params != region2_params:
        raise ValueError('Inputs do not have the same parameters')

    assert region1.meta == region2.meta
    assert region1.visual == region2.visual

    # now check the parameter values
    # Note that Quantity comparisons allow for different units
    # if they directly convertible (e.g., 1. * u.deg == 60. * u.arcmin)
    for param in region1_params:
        value1 = getattr(region1, param)
        value2 = getattr(region2, param)
        if isinstance(value1, SkyCoord):
            assert_skycoord_allclose(value1, value2, **kwargs)
        elif isinstance(value1, PixCoord):
            assert_allclose(value1.xy, value2.xy)
        elif isinstance(value1, u.Quantity):
            assert_quantity_allclose(value1, value2, **kwargs)
        elif isinstance(value1, str):
            assert value1 == value2
        else:
            assert_allclose(value1, value2)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines text regions in both pixel and sky coordinates.
"""

from regions._utils.wcs_helpers import pixel_scale_angle_at_skycoord
from regions.core.attributes import (RegionMetaDescr, RegionVisualDescr,
                                     ScalarPixCoord, ScalarSkyCoord)
from regions.shapes.point import PointPixelRegion, PointSkyRegion

__all__ = ['TextSkyRegion', 'TextPixelRegion']


class TextPixelRegion(PointPixelRegion):
    """
    A text string in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The leftmost point of the text string before rotation.
    text : str
        The text string.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from regions import PixCoord, TextPixelRegion, RegionVisual
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1)

        center = PixCoord(x=15, y=10)
        visual = RegionVisual({'textangle': 30})
        reg = TextPixelRegion(center=center, text="Hello World!",
                              visual=visual)
        reg.plot(ax=ax)

        ax.set_xlim(10, 30)
        ax.set_ylim(2.5, 20)
        ax.set_aspect('equal')
    """

    _params = ('center', 'text')
    _mpl_artist = 'Text'
    center = ScalarPixCoord('The leftmost pixel position (before rotation) '
                            'as a |PixCoord|.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, center, text, meta=None, visual=None):
        super().__init__(center, meta, visual)
        self.text = text

    def to_sky(self, wcs):
        center = wcs.pixel_to_world(self.center.x, self.center.y)

        # rotation value is relative to the coordinate system axes;
        # convert from counterclockwise angle from the positive x axis
        # to angle relative to WCS longitude axis
        visual = self.visual
        if 'rotation' in self.visual:
            _, _, angle = pixel_scale_angle_at_skycoord(center, wcs)
            visual = visual.copy()
            visual['rotation'] -= angle.to('deg').value - 90.

        return TextSkyRegion(center, self.text, meta=self.meta.copy(),
                             visual=visual.copy())

    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Return a matplotlib Text object for this region
        (`matplotlib.text.Text`).

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        **kwargs : dict
            Any keyword arguments accepted by `~matplotlib.text.Text`.
            These keywords will override any visual meta attributes of
            this region.

        Returns
        -------
        artist : `~matplotlib.text.Text`
            A matplotlib Text object.
        """
        from matplotlib.text import Text

        mpl_kwargs = self.visual.define_mpl_kwargs(self._mpl_artist)
        mpl_kwargs.update(kwargs)

        return Text(self.center.x - origin[0], self.center.y - origin[1],
                    self.text, **mpl_kwargs)


class TextSkyRegion(PointSkyRegion):
    """
    A text string in sky coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The leftmost position of the text string before rotation.
    text : str
        The text string.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.
    """

    _params = ('center', 'text')
    center = ScalarSkyCoord('The leftmost position (before rotation) as a '
                            '|SkyCoord|.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, center, text, meta=None, visual=None):
        super().__init__(center, meta, visual)
        self.text = text

    def to_pixel(self, wcs):
        center, _, angle = pixel_scale_angle_at_skycoord(self.center, wcs)

        # rotation value is relative to the WCS longitude axis;
        # convert to counterclockwise angle from the positive x axis
        visual = self.visual
        if 'rotation' in self.visual:
            visual = visual.copy()
            visual['rotation'] += angle.to('deg').value - 90.

        return TextPixelRegion(center, self.text, meta=self.meta.copy(),
                               visual=visual.copy())



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.data import get_pkg_data_filename
from astropy.wcs import WCS
from numpy.testing import assert_allclose

from regions._utils.optional_deps import HAS_MATPLOTLIB
from regions.core import PixCoord, RegionMeta, RegionVisual
from regions.shapes.line import LinePixelRegion, LineSkyRegion
from regions.shapes.tests.test_common import (BaseTestPixelRegion,
                                              BaseTestSkyRegion)
from regions.tests.helpers import make_simple_wcs


@pytest.fixture(scope='session', name='wcs')
def wcs_fixture():
    filename = get_pkg_data_filename('data/example_header.fits')
    header = fits.getheader(filename)
    return WCS(header)


class TestLinePixelRegion(BaseTestPixelRegion):
    meta = RegionMeta({'text': 'test'})
    visual = RegionVisual({'color': 'blue'})
    reg = LinePixelRegion(PixCoord(3, 4), PixCoord(4, 4), meta=meta,
                          visual=visual)
    sample_box = [-2, 8, -1, 9]
    inside = []
    outside = [(3.1, 4.2), (5, 4)]
    expected_area = 0
    expected_repr = ('<LinePixelRegion(start=PixCoord(x=3, y=4), '
                     'end=PixCoord(x=4, y=4))>')
    expected_str = ('Region: LinePixelRegion\nstart: PixCoord(x=3, y=4)\n'
                    'end: PixCoord(x=4, y=4)')

    def test_copy(self):
        reg = self.reg.copy()
        assert reg.start.xy == (3, 4)
        assert reg.end.xy == (4, 4)
        assert reg.meta == self.meta
        assert reg.visual == self.visual

    def test_pix_sky_roundtrip(self):
        wcs = make_simple_wcs(SkyCoord(2 * u.deg, 3 * u.deg), 0.1 * u.deg, 20)
        reg_new = self.reg.to_sky(wcs).to_pixel(wcs)
        assert_allclose(reg_new.start.x, self.reg.start.x)
        assert_allclose(reg_new.start.y, self.reg.start.y)
        assert_allclose(reg_new.end.x, self.reg.end.x)
        assert_allclose(reg_new.end.y, self.reg.end.y)
        assert reg_new.meta == self.reg.meta
        assert reg_new.visual == self.reg.visual

        # test that converted region meta and visual are copies and not views
        reg_new.meta['text'] = 'new'
        reg_new.visual['color'] = 'green'
        assert reg_new.meta['text'] != self.reg.meta['text']
        assert reg_new.visual['color'] != self.reg.visual['color']

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_as_artist(self):
        patch = self.reg.as_artist()
        assert 'Arrow' in str(patch)

    def test_rotate(self):
        reg = self.reg.rotate(PixCoord(2, 3), 90 * u.deg)
        assert_allclose(reg.start.xy, (1, 4))
        assert_allclose(reg.end.xy, (1, 5))

    def test_eq(self):
        reg = self.reg.copy()
        assert reg == self.reg
        reg.start = PixCoord(1, 2)
        assert reg != self.reg


class TestLineSkyRegion(BaseTestSkyRegion):
    meta = RegionMeta({'text': 'test'})
    visual = RegionVisual({'color': 'blue'})
    start = SkyCoord(3 * u.deg, 4 * u.deg, frame='galactic')
    end = SkyCoord(3 * u.deg, 5 * u.deg, frame='galactic')
    reg = LineSkyRegion(start, end, meta=meta, visual=visual)

    expected_repr = ('<LineSkyRegion(start=<SkyCoord (Galactic): (l, b) '
                     'in deg\n    (3., 4.)>, end=<SkyCoord (Galactic): '
                     '(l, b) in deg\n    (3., 5.)>)>')
    expected_str = ('Region: LineSkyRegion\nstart: <SkyCoord (Galactic): '
                    '(l, b) in deg\n    (3., 4.)>\nend: <SkyCoord '
                    '(Galactic): (l, b) in deg\n    (3., 5.)>')

    def test_copy(self):
        reg = self.reg.copy()
        assert_allclose(reg.start.b.deg, 4)
        assert_allclose(reg.end.b.deg, 5)
        assert reg.meta == self.meta
        assert reg.visual == self.visual

    def test_transformation(self, wcs):
        pixline = self.reg.to_pixel(wcs)

        assert_allclose(pixline.start.x, -50.5)
        assert_allclose(pixline.start.y, 299.5)
        assert_allclose(pixline.end.x, -50.5)
        assert_allclose(pixline.end.y, 349.5)

        skyline = pixline.to_sky(wcs)

        assert_quantity_allclose(skyline.start.data.lon,
                                 self.reg.start.data.lon)
        assert_quantity_allclose(skyline.start.data.lat,
                                 self.reg.start.data.lat)
        assert_quantity_allclose(skyline.end.data.lon, self.reg.end.data.lon)
        assert_quantity_allclose(skyline.end.data.lat, self.reg.end.data.lat)

    def test_contains(self, wcs):
        position = SkyCoord([1, 2] * u.deg, [3, 4] * u.deg)
        # lines do not contain things
        assert all(self.reg.contains(position, wcs)
                   == np.array([False, False], dtype='bool'))

    def test_eq(self):
        reg = self.reg.copy()
        assert reg == self.reg
        reg.start = SkyCoord(1 * u.deg, 2 * u.deg, frame='galactic')
        assert reg != self.reg
        reg.start = SkyCoord(3 * u.deg, 4 * u.deg, frame='icrs')
        assert reg != self.reg



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
The tests in this file simply check what functionality is currently
implemented and doesn't check anything about correctness.
"""
import itertools
from collections import OrderedDict

import astropy.units as u
import pytest
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS

from regions.core.core import PixelRegion, Region, SkyRegion
from regions.core.mask import RegionMask
from regions.core.metadata import RegionMeta, RegionVisual
from regions.core.pixcoord import PixCoord
from regions.shapes.annulus import (CircleAnnulusPixelRegion,
                                    CircleAnnulusSkyRegion,
                                    EllipseAnnulusPixelRegion,
                                    EllipseAnnulusSkyRegion,
                                    RectangleAnnulusPixelRegion,
                                    RectangleAnnulusSkyRegion)
from regions.shapes.circle import CirclePixelRegion, CircleSkyRegion
from regions.shapes.ellipse import EllipsePixelRegion, EllipseSkyRegion
from regions.shapes.point import PointPixelRegion, PointSkyRegion
from regions.shapes.polygon import PolygonPixelRegion, PolygonSkyRegion
from regions.shapes.rectangle import RectanglePixelRegion, RectangleSkyRegion

PIXEL_REGIONS = [
    CirclePixelRegion(PixCoord(3, 4), radius=5),
    CircleAnnulusPixelRegion(PixCoord(3, 4), 5, 7),
    EllipsePixelRegion(PixCoord(3, 4), width=7, height=5, angle=3 * u.deg),
    EllipseAnnulusPixelRegion(PixCoord(3, 4), 7, 9, 8, 10, angle=3 * u.deg),
    PolygonPixelRegion(PixCoord([1, 4, 3], [2, 4, 4])),
    RectanglePixelRegion(PixCoord(6, 5), width=3, height=5),
    RectangleAnnulusPixelRegion(PixCoord(6, 5), 3, 6, 5, 7),
    PointPixelRegion(PixCoord(1, 2))]

SKY_REGIONS = [
    CircleSkyRegion(SkyCoord(3 * u.deg, 4 * u.deg), radius=5 * u.deg),
    CircleAnnulusSkyRegion(SkyCoord(3 * u.deg, 4 * u.deg), 5 * u.deg,
                           7 * u.deg),
    EllipseSkyRegion(SkyCoord(3 * u.deg, 4 * u.deg), width=7 * u.deg,
                     height=5 * u.deg, angle=3 * u.deg),
    EllipseAnnulusSkyRegion(SkyCoord(3 * u.deg, 4 * u.deg), 7 * u.deg,
                            9 * u.deg, 5 * u.deg, 9 * u.deg, angle=3 * u.deg),
    PolygonSkyRegion(SkyCoord([1, 4, 3] * u.deg, [2, 4, 4] * u.deg)),
    RectangleSkyRegion(SkyCoord(6 * u.deg, 5 * u.deg), width=3 * u.deg,
                       height=5 * u.deg),
    RectangleAnnulusSkyRegion(SkyCoord(6 * u.deg, 5 * u.deg), 3 * u.deg,
                              5 * u.deg, 5 * u.deg, 7 * u.deg),
    PointSkyRegion(SkyCoord(6 * u.deg, 5 * u.deg))]

MASK_MODES = ['center', 'exact', 'subpixels']
COMMON_WCS = WCS(naxis=2)
COMMON_WCS.wcs.ctype = 'RA---TAN', 'DEC--TAN'


def ids_func(arg):
    if isinstance(arg, Region):
        return arg.__class__.__name__
    else:
        return str(arg)


@pytest.mark.parametrize('region', [PIXEL_REGIONS[i] for i in (0, 2, 5)],
                         ids=ids_func)
def test_pix_in(region):
    assert PixCoord(5, 4) in region


@pytest.mark.parametrize('region', PIXEL_REGIONS, ids=ids_func)
def test_pix_area(region):
    # TODO: remove the pytest.skip once polygon area is implemented
    if isinstance(region, PolygonPixelRegion):
        pytest.skip()

    area = region.area
    assert not isinstance(area, u.Quantity)


@pytest.mark.parametrize('region', PIXEL_REGIONS, ids=ids_func)
def test_pix_to_sky(region):
    try:
        sky_region = region.to_sky(COMMON_WCS)
        assert isinstance(sky_region, SkyRegion)
    except NotImplementedError:
        pytest.xfail()


@pytest.mark.parametrize(('region', 'mode'),
                         itertools.product(PIXEL_REGIONS, MASK_MODES),
                         ids=ids_func)
def test_pix_to_mask(region, mode):
    try:
        mask = region.to_mask(mode=mode)
        assert isinstance(mask, RegionMask)
    except NotImplementedError:
        pytest.xfail()


@pytest.mark.parametrize('region', SKY_REGIONS, ids=ids_func)
def test_sky_in(region):
    region.contains(SkyCoord(1 * u.deg, 1 * u.deg, frame='icrs'), COMMON_WCS)


@pytest.mark.parametrize('region', SKY_REGIONS, ids=ids_func)
def test_sky_in_array(region):
    region.contains(SkyCoord([1, 2, 3] * u.deg, [3, 2, 1] * u.deg,
                             frame='icrs'), COMMON_WCS)


@pytest.mark.parametrize('region', SKY_REGIONS, ids=ids_func)
def test_sky_to_pix(region):
    pix_region = region.to_pixel(wcs=COMMON_WCS)
    assert isinstance(pix_region, PixelRegion)


@pytest.mark.parametrize('region', PIXEL_REGIONS, ids=ids_func)
def test_attribute_validation_pixel_regions(region):
    invalid_values = dict(center=[PixCoord([1, 2], [2, 3]), 1,
                                  SkyCoord(1 * u.deg, 2 * u.deg),
                                  (10, 10), (10 * u.deg, 10 * u.deg)],
                          radius=[u.Quantity('1deg'), [1], PixCoord(1, 2),
                                  3 * u.km, 0.0, -10.],
                          angle=[u.Quantity([1 * u.deg, 2 * u.deg]), 2,
                                 PixCoord(1, 2), 3 * u.km],
                          vertices=[u.Quantity('1'), 2, PixCoord(1, 2),
                                    PixCoord([[1, 2]], [[2, 3]]), 3 * u.km,
                                    (10, 10), (10 * u.deg, 10 * u.deg)])
    invalid_values['width'] = invalid_values['radius']
    invalid_values['height'] = invalid_values['radius']
    invalid_values['inner_height'] = invalid_values['radius']
    invalid_values['inner_width'] = invalid_values['radius']
    invalid_values['outer_height'] = invalid_values['radius']
    invalid_values['outer_width'] = invalid_values['radius']
    invalid_values['inner_radius'] = invalid_values['radius']
    invalid_values['outer_radius'] = invalid_values['radius']
    invalid_values['start'] = invalid_values['center']
    invalid_values['end'] = invalid_values['radius']

    for attr in invalid_values:
        if hasattr(region, attr):
            for val in invalid_values.get(attr, None):
                with pytest.raises(ValueError) as excinfo:
                    setattr(region, attr, val)
                assert f'{attr!r} must' in str(excinfo.value)


@pytest.mark.parametrize('region', SKY_REGIONS, ids=ids_func)
def test_attribute_validation_sky_regions(region):
    invalid_values = dict(center=[PixCoord([1, 2], [2, 3]), 1,
                                  SkyCoord([1 * u.deg], [2 * u.deg]),
                                  (10, 10), (10 * u.deg, 10 * u.deg)],
                          radius=[u.Quantity([1 * u.deg, 5 * u.deg]),
                                  [1], SkyCoord(1 * u.deg, 2 * u.deg),
                                  1, 3 * u.km, 0.0 * u.deg, -10. * u.deg],
                          angle=[u.Quantity([1 * u.deg, 2 * u.deg]), 2,
                                 SkyCoord(1 * u.deg, 2 * u.deg), 3. * u.km],
                          vertices=[u.Quantity('1deg'), 2,
                                    SkyCoord(1 * u.deg, 2 * u.deg),
                                    SkyCoord([[1 * u.deg, 2 * u.deg]],
                                             [[2 * u.deg, 3 * u.deg]]),
                                    3 * u.km, (10, 10),
                                    (10 * u.deg, 10 * u.deg)])

    invalid_values['width'] = invalid_values['radius']
    invalid_values['height'] = invalid_values['radius']
    invalid_values['inner_height'] = invalid_values['radius']
    invalid_values['inner_width'] = invalid_values['radius']
    invalid_values['outer_height'] = invalid_values['radius']
    invalid_values['outer_width'] = invalid_values['radius']
    invalid_values['inner_radius'] = invalid_values['radius']
    invalid_values['outer_radius'] = invalid_values['radius']
    invalid_values['start'] = invalid_values['center']
    invalid_values['end'] = invalid_values['radius']

    for attr in invalid_values:
        if hasattr(region, attr):
            for val in invalid_values.get(attr, None):
                with pytest.raises(ValueError) as excinfo:
                    setattr(region, attr, val)
                assert f'{attr!r} must' in str(excinfo.value)


@pytest.mark.parametrize('region', PIXEL_REGIONS + SKY_REGIONS, ids=ids_func)
def test_metadata(region):
    region.meta = {'text': 'hello'}
    region.visual = {'color': 'blue'}
    assert isinstance(region.meta, RegionMeta)
    assert isinstance(region.visual, RegionVisual)

    # dict subclasses are allowed
    region.meta = OrderedDict({'text': 'hello'})
    region.visual = OrderedDict({'color': 'blue'})
    assert isinstance(region.meta, RegionMeta)
    assert isinstance(region.visual, RegionVisual)

    with pytest.raises(ValueError):
        region.meta = 1
    with pytest.raises(ValueError):
        region.visual = 'blue'



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines circular regions in both pixel and sky coordinates.
"""

import math

import astropy.units as u
import numpy as np
from astropy.coordinates import Angle

from regions._geometry import circular_overlap_grid
from regions._utils.wcs_helpers import pixel_scale_angle_at_skycoord
from regions.core.attributes import (PositiveScalar, PositiveScalarAngle,
                                     RegionMetaDescr, RegionVisualDescr,
                                     ScalarPixCoord, ScalarSkyCoord)
from regions.core.bounding_box import RegionBoundingBox
from regions.core.core import PixelRegion, SkyRegion
from regions.core.mask import RegionMask
from regions.core.metadata import RegionMeta, RegionVisual
from regions.core.pixcoord import PixCoord

__all__ = ['CirclePixelRegion', 'CircleSkyRegion']


class CirclePixelRegion(PixelRegion):
    """
    A circle defined using pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The center position.
    radius : float
        The radius in pixels.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from regions import PixCoord, CirclePixelRegion
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1)

        reg = CirclePixelRegion(PixCoord(x=8, y=7), radius=3.5)
        patch = reg.plot(ax=ax, facecolor='none', edgecolor='red', lw=2,
                         label='Circle')

        ax.legend(handles=(patch,), loc='upper center')
        ax.set_xlim(0, 15)
        ax.set_ylim(0, 15)
        ax.set_aspect('equal')
    """

    _params = ('center', 'radius')
    _mpl_artist = 'Patch'
    center = ScalarPixCoord('The center pixel position as a |PixCoord|.')
    radius = PositiveScalar('The radius in pixels as a float.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, center, radius, meta=None, visual=None):
        self.center = center
        self.radius = radius
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    @property
    def area(self):
        return math.pi * self.radius ** 2

    def contains(self, pixcoord):
        pixcoord = PixCoord._validate(pixcoord, name='pixcoord')
        in_circle = self.center.separation(pixcoord) < self.radius
        if self.meta.get('include', True):
            return in_circle
        else:
            return np.logical_not(in_circle)

    def to_sky(self, wcs):
        # TODO: write a pixel_to_skycoord_scale_angle
        center = wcs.pixel_to_world(self.center.x, self.center.y)
        _, pixscale, _ = pixel_scale_angle_at_skycoord(center, wcs)
        radius = Angle(self.radius * u.pix * pixscale, 'arcsec')
        return CircleSkyRegion(center, radius, meta=self.meta.copy(),
                               visual=self.visual.copy())

    @property
    def bounding_box(self):
        """
        Bounding box (`~regions.RegionBoundingBox`).
        """
        xmin = self.center.x - self.radius
        xmax = self.center.x + self.radius
        ymin = self.center.y - self.radius
        ymax = self.center.y + self.radius

        return RegionBoundingBox.from_float(xmin, xmax, ymin, ymax)

    def to_mask(self, mode='center', subpixels=1):
        self._validate_mode(mode, subpixels)

        if mode == 'center':
            mode = 'subpixels'
            subpixels = 1

        # Find bounding box and mask size
        bbox = self.bounding_box
        ny, nx = bbox.shape

        # Find position of pixel edges and recenter so that circle is at
        # origin
        xmin = float(bbox.ixmin) - 0.5 - self.center.x
        xmax = float(bbox.ixmax) - 0.5 - self.center.x
        ymin = float(bbox.iymin) - 0.5 - self.center.y
        ymax = float(bbox.iymax) - 0.5 - self.center.y

        use_exact = 0 if mode == 'subpixels' else 1

        fraction = circular_overlap_grid(xmin, xmax, ymin, ymax, nx, ny,
                                         self.radius, use_exact, subpixels)

        return RegionMask(fraction, bbox=bbox)

    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Return a matplotlib patch object for this region
        (`matplotlib.patches.Circle`).

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        **kwargs : dict
            Any keyword arguments accepted by
            `~matplotlib.patches.Circle`. These keywords will override
            any visual meta attributes of this region.

        Returns
        -------
        artist : `~matplotlib.patches.Circle`
            A matplotlib circle patch.
        """
        from matplotlib.patches import Circle

        xy = self.center.x - origin[0], self.center.y - origin[1]
        radius = self.radius
        mpl_kwargs = self.visual.define_mpl_kwargs(self._mpl_artist)
        mpl_kwargs.update(kwargs)

        return Circle(xy=xy, radius=radius, **mpl_kwargs)

    def rotate(self, center, angle):
        """
        Rotate the region.

        Positive ``angle`` corresponds to counter-clockwise rotation.

        Parameters
        ----------
        center : `~regions.PixCoord`
            The rotation center point.
        angle : `~astropy.coordinates.Angle`
            The rotation angle.

        Returns
        -------
        region : `CirclePixelRegion`
            The rotated region (which is an independent copy).
        """
        center = self.center.rotate(center, angle)
        return self.copy(center=center)


class CircleSkyRegion(SkyRegion):
    """
    A circle defined using sky coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The center position.
    radius : `~astropy.units.Quantity`
        The radius in angular units.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.
    """

    _params = ('center', 'radius')
    center = ScalarSkyCoord('The center position as a |SkyCoord|.')
    radius = PositiveScalarAngle('The radius as a |Quantity| angle.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, center, radius, meta=None, visual=None):
        self.center = center
        self.radius = radius
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    def to_pixel(self, wcs):
        center, pixscale, _ = pixel_scale_angle_at_skycoord(self.center, wcs)
        radius = (self.radius / pixscale).to(u.pix).value
        return CirclePixelRegion(center, radius, meta=self.meta.copy(),
                                 visual=self.visual.copy())



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings

import astropy.units as u
import numpy as np
from astropy.io import fits
from astropy.table import QTable
from astropy.utils.exceptions import AstropyUserWarning

from regions.core import PixCoord, RegionMeta, Regions
from regions.core.registry import RegionsRegistry
from regions.io.fits.core import FITSParserError, shape_map

__all__ = []


@RegionsRegistry.register(Regions, 'parse', 'fits')
def _parse_fits(region_table):
    """
    Parse a FITS region table.

    Parameters
    ----------
    region_table : `~astropy.table.Table`
        The table contents of a FITS region file.

    Returns
    -------
    regions : `regions.Regions`
        A `Regions` object containing a list of `~regions.Region`
        objects.
    """
    regions = parse_table(region_table)
    return Regions(regions)


@RegionsRegistry.register(Regions, 'read', 'fits')
def _read_fits(filename, cache=False):
    """
    Read a FITS region file, converting a FITS regions table to a list
    of `~regions.Region` objects.

    Parameters
    ----------
    filename : str
        The FITS region filename. The first "REGION" FITS extension will
        be used.

    cache : bool or 'update', optional
        Whether to cache the contents of remote URLs. If 'update', check
        the remote URL for a new version but store the result in the
        cache.

    Returns
    -------
    regions : list
        A list of `~regions.Region` objects.
    """
    with fits.open(filename, cache=cache) as hdul:
        for hdu in hdul:
            # use the first 'REGION' HDU
            if hdu.name == 'REGION':
                region_table = QTable.read(hdu)
                regions = _parse_fits(region_table)
                return regions

        raise FITSParserError('An extension with name (EXTNAME) "REGION" '
                              'was not found')


def get_shape(region_row):
    include = 1
    shape_key = 'SHAPE'
    if shape_key not in region_row.colnames:
        shape = 'point'
        return shape, include

    shape = region_row[shape_key].lower()
    if shape[0] == '!':
        include = 0
        shape = shape[1:]

    supported_shapes = list(shape_map.keys())
    unsupported_shapes = ['pie', 'sector', 'diamond', 'rhombus',
                          'rotdiamond', 'rotrhombus']
    valid_shapes = supported_shapes + unsupported_shapes

    if shape not in valid_shapes:
        raise FITSParserError(f'{shape!r} is not a valid FITS region shape')
    if shape not in supported_shapes:
        warnings.warn(f'{shape!r} is not supported by the regions package, '
                      'skipping.', AstropyUserWarning)
        shape = None

    return shape, include


def get_column_values(region_row, colname):
    index = None
    if colname[-1].isdigit():
        index = int(colname[-1])
        colname = colname[:-1]

    value = np.atleast_1d(region_row[colname])
    if isinstance(value, u.Quantity) and value.unit == u.pixel:
        value = value.value  # remove pixel units

    if index is None:  # polygon uses all values
        return value

    try:
        return value[index]
    except IndexError as exc:
        raise FITSParserError(f'The {colname!r} column must have more '
                              f'than {index!r} values for the '
                              'region.') from exc


def get_shape_params(shape, region_row, shape_columns):
    values = [get_column_values(region_row, column)
              for column in shape_columns]

    if 'rectangle' in shape:
        (xmin, xmax, ymin, ymax) = values[0:4]
        xcenter = 0.5 * (xmin + xmax)
        ycenter = 0.5 * (ymin + ymax)
        xsize = xmax - xmin
        ysize = ymax - ymin

        shape_params = [PixCoord(xcenter, ycenter), xsize, ysize]
        if shape == 'rotrectangle':
            shape_params.append(values[-1])  # angle

        return shape_params

    # center (or polygon) coordinates for all other regions
    shape_params = [PixCoord(values[0], values[1])]

    # shape params
    if shape == 'ellipse':
        # FITS uses semi-axis lengths;
        # the last value is always the rotation angle
        values[2:-1] = list(np.array(values[2:-1]) * 2.)

    shape_params.extend(values[2:])

    return shape_params


def parse_row(region_row):
    shape, include = get_shape(region_row)
    if shape is None:
        return None

    region_cls, shape_columns = shape_map[shape]

    for column in shape_columns:
        if column[-1].isdigit():
            column = column[:-1]
        if column not in region_row.colnames:
            warnings.warn(f'Table columns are missing for {shape!r} shape, '
                          'skipping.', AstropyUserWarning)
            return None

    shape_params = get_shape_params(shape, region_row, shape_columns)
    region = region_cls(*shape_params)

    meta = {}
    if include == 0:
        meta = {'include': include}
        region.meta = RegionMeta(meta)

    shape_key = 'COMPONENT'
    if shape_key in region_row.colnames:
        component = int(region_row[shape_key])
        meta = {'component': component}

    if meta:
        region.meta = RegionMeta(meta)

    return region


def parse_table(region_table):
    valid_columns = ('X', 'Y', 'SHAPE', 'R', 'ROTANG', 'COMPONENT')

    for column in region_table.colnames:
        if column not in valid_columns:
            raise FITSParserError(f'{column!r} is not a valid column name')

    regions = []
    for row in region_table:
        region = parse_row(row)
        if region is not None:
            regions.append(region)

    return regions



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from astropy.wcs import WCS
from numpy.testing import assert_allclose

from regions.core import PixCoord, RegionMeta, RegionVisual
from regions.shapes.tests.test_common import (BaseTestPixelRegion,
                                              BaseTestSkyRegion)
from regions.shapes.text import TextPixelRegion, TextSkyRegion
from regions.tests.helpers import make_simple_wcs


@pytest.fixture(scope='session', name='wcs')
def wcs_fixture():
    filename = get_pkg_data_filename('data/example_header.fits')
    header = fits.getheader(filename)
    return WCS(header)


class TestTextPixelRegion(BaseTestPixelRegion):
    meta = RegionMeta({'text': 'test'})
    visual = RegionVisual({'color': 'blue'})
    reg = TextPixelRegion(PixCoord(3, 4), 'Sample Text', meta=meta,
                          visual=visual)
    sample_box = [-2, 8, -1, 9]
    inside = []
    outside = [(3.1, 4.2), (5, 4)]
    expected_area = 0
    expected_repr = ('<TextPixelRegion(center=PixCoord(x=3, y=4), '
                     "text='Sample Text')>")
    expected_str = ('Region: TextPixelRegion\ncenter: PixCoord(x=3, y=4)\n'
                    "text: 'Sample Text'")

    def test_copy(self):
        reg = self.reg.copy()
        assert reg.center.xy == (3, 4)
        assert reg.text == 'Sample Text'
        assert reg.meta == self.meta
        assert reg.visual == self.visual

    def test_pix_sky_roundtrip(self):
        wcs = make_simple_wcs(SkyCoord(2 * u.deg, 3 * u.deg), 0.1 * u.deg, 20)
        reg_new = self.reg.to_sky(wcs).to_pixel(wcs)
        assert_allclose(reg_new.center.x, self.reg.center.x)
        assert_allclose(reg_new.center.y, self.reg.center.y)
        assert reg_new.meta == self.reg.meta
        assert reg_new.visual == self.reg.visual

        # test that converted region meta and visual are copies and not views
        reg_new.meta['text'] = 'new'
        reg_new.visual['color'] = 'green'
        assert reg_new.meta['text'] != self.reg.meta['text']
        assert reg_new.visual['color'] != self.reg.visual['color']

    def test_rotate(self):
        reg = self.reg.rotate(PixCoord(2, 3), 90 * u.deg)
        assert_allclose(reg.center.xy, (1, 4))

    def test_eq(self):
        reg = self.reg.copy()
        assert reg == self.reg
        reg.text = 'Hello World!'
        assert reg != self.reg


class TestTextSkyRegion(BaseTestSkyRegion):
    meta = RegionMeta({'text': 'test'})
    visual = RegionVisual({'color': 'blue'})
    reg = TextSkyRegion(SkyCoord(3, 4, unit='deg'), 'Sample Text',
                        meta=meta, visual=visual)
    expected_repr = ('<TextSkyRegion(center=<SkyCoord (ICRS): (ra, dec) in '
                     "deg\n    (3., 4.)>, text='Sample Text')>")
    expected_str = ('Region: TextSkyRegion\ncenter: <SkyCoord (ICRS): '
                    "(ra, dec) in deg\n    (3., 4.)>\ntext: 'Sample Text'")

    def test_copy(self):
        reg = self.reg.copy()
        assert_allclose(reg.center.ra.deg, 3)
        assert reg.text == 'Sample Text'
        assert reg.meta == self.meta
        assert reg.visual == self.visual

    def test_contains(self, wcs):
        position = SkyCoord([1, 2] * u.deg, [3, 4] * u.deg)
        # Nothing is in a text region
        assert all(self.reg.contains(position, wcs)
                   == np.array([False, False], dtype='bool'))

    def test_eq(self):
        reg = self.reg.copy()
        assert reg == self.reg
        reg.text = 'Hello World!'
        assert reg != self.reg



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest
from astropy.tests.helper import assert_quantity_allclose
from astropy.units.quantity import Quantity

from regions.io.crtf.io_core import _to_shape_list
from regions.io.crtf.read import _CRTFParser


def test_shape_crtf():
    reg_str = ('circle[[18h12m24s, -23d11m00s], 2.3arcsec], linewidth=2, '
               'coord=J2000, symsize=2')

    parser = _CRTFParser(reg_str)
    shape1 = parser.shapes[0]
    region = parser.shapes.to_regions()
    shape2 = _to_shape_list(region, 'fk5')[0]

    assert shape1.coordsys == shape2.coordsys
    assert shape1.region_type == shape2.region_type
    assert shape1.include == shape2.include
    assert_quantity_allclose(shape1.coord, shape2.coord)
    # assert dict(shape1.meta) == dict(shape2.meta)
    assert shape1.composite == shape2.composite

    # Pixel origin is assumed to be 0 origin. Needs to be checked though.
    reg_str = 'circle[[1pix, 2pix], 3pix], linewidth=2, coord=image, symsize=2'

    parser = _CRTFParser(reg_str)
    shape = parser.shapes[0]
    assert_quantity_allclose(shape.coord,
                             [Quantity('1'), Quantity('2'), Quantity('3')])


def test_valid_shape():
    reg_str = ('circle[[18h12m24s, -23d11m00s], 2.3arcsec], linewidth=2, '
               'coord=J2000, symsize=2')

    shape = _CRTFParser(reg_str).shapes[0]

    with pytest.raises(ValueError) as excinfo:
        shape.region_type = 'box'

    assert '"box" is not a valid region type' in str(excinfo.value)

    shape = _CRTFParser(reg_str).shapes[0]
    with pytest.raises(ValueError) as excinfo:
        shape.coordsys = 'hello'

    estr = '"hello" is not a valid coordinate reference frame'
    assert estr in str(excinfo.value)


def test_valid_ellipse_crtf():
    reg_str1 = 'ellipse[[1pix, 2pix], [3pix, 4pix], 5deg]'
    reg_str2 = 'ellipse[[1deg, 2deg], [3deg, 4deg], 5deg], coord=fk5'

    shape1 = _CRTFParser(reg_str1, 'warn').shapes[0]
    shape2 = _CRTFParser(reg_str2, 'warn').shapes[0]

    assert_quantity_allclose(shape1.coord[:2], [1, 2])
    assert_quantity_allclose(shape2.coord[:2],
                             [Quantity('1deg'), Quantity('2deg')])

    assert_quantity_allclose(shape1.coord[2:-1], [8, 6])
    assert_quantity_allclose(shape2.coord[2:-1],
                             [Quantity('8deg'), Quantity('6deg')])

    assert_quantity_allclose(shape1.coord[-1], Quantity('5deg'))
    assert_quantity_allclose(shape2.coord[-1], Quantity('5deg'))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines annulus regions in both pixel and sky coordinates.
"""

import abc
import operator

import astropy.units as u

from regions._utils.wcs_helpers import pixel_scale_angle_at_skycoord
from regions.core.attributes import (PositiveScalar, PositiveScalarAngle,
                                     RegionMetaDescr, RegionVisualDescr,
                                     ScalarAngle, ScalarPixCoord,
                                     ScalarSkyCoord)
from regions.core.compound import CompoundPixelRegion
from regions.core.core import PixelRegion, SkyRegion
from regions.core.metadata import RegionMeta, RegionVisual
from regions.core.pixcoord import PixCoord
from regions.shapes.circle import CirclePixelRegion
from regions.shapes.ellipse import EllipsePixelRegion, EllipseSkyRegion
from regions.shapes.rectangle import RectanglePixelRegion, RectangleSkyRegion

__all__ = ['AnnulusPixelRegion', 'AsymmetricAnnulusPixelRegion',
           'AsymmetricAnnulusSkyRegion',
           'CircleAnnulusPixelRegion', 'CircleAnnulusSkyRegion',
           'EllipseAnnulusPixelRegion', 'EllipseAnnulusSkyRegion',
           'RectangleAnnulusPixelRegion', 'RectangleAnnulusSkyRegion']


class AnnulusPixelRegion(PixelRegion, abc.ABC):
    """
    A base class for annulus pixel regions.
    """

    @property
    def _compound_region(self):
        return CompoundPixelRegion(self._inner_region, self._outer_region,
                                   operator.xor, self.meta, self.visual)

    @property
    def area(self):
        return self._outer_region.area - self._inner_region.area

    @property
    def bounding_box(self):
        return self._outer_region.bounding_box

    def contains(self, pixcoord):
        return self._compound_region.contains(pixcoord)

    def as_artist(self, origin=(0, 0), **kwargs):
        return self._compound_region.as_artist(origin, **kwargs)

    def to_mask(self, mode='center', subpixels=5):
        return self._compound_region.to_mask(mode, subpixels)

    def rotate(self, center, angle):
        """
        Rotate the region.

        Positive ``angle`` corresponds to counter-clockwise rotation.

        Parameters
        ----------
        center : `~regions.PixCoord`
            The rotation center point.
        angle : `~astropy.coordinates.Angle`
            The rotation angle.

        Returns
        -------
        region : `~regions.AnnulusPixelRegion`
            The rotated region (which is an independent copy).
        """
        changes = {}
        changes['center'] = self.center.rotate(center, angle)
        if hasattr(self, 'angle'):
            changes['angle'] = self.angle + angle
        return self.copy(**changes)


class CircleAnnulusPixelRegion(AnnulusPixelRegion):
    """
    A circular annulus in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the annulus.
    inner_radius : float
        The inner radius of the annulus in pixels.
    outer_radius : float
        The outer radius of the annulus in pixels.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from regions import PixCoord, CircleAnnulusPixelRegion
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1)

        reg = CircleAnnulusPixelRegion(PixCoord(x=6, y=6),
                                       inner_radius=5.5,
                                       outer_radius=8.0)
        patch = reg.plot(ax=ax, facecolor='none', edgecolor='red', lw=2,
                         label='Circle Annulus')

        ax.legend(handles=(patch,), loc='upper center')
        ax.set_xlim(-5, 20)
        ax.set_ylim(-5, 20)
        ax.set_aspect('equal')
    """

    _component_class = CirclePixelRegion
    _params = ('center', 'inner_radius', 'outer_radius')
    center = ScalarPixCoord('The center pixel position as a |PixCoord|.')
    inner_radius = PositiveScalar('The inner radius in pixels as a float.')
    outer_radius = PositiveScalar('The outer radius in pixels as a float.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, center, inner_radius, outer_radius, meta=None,
                 visual=None):
        self.center = center
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

        if inner_radius >= outer_radius:
            raise ValueError('outer_radius must be greater than inner_radius')

    @property
    def _inner_region(self):
        return self._component_class(self.center, self.inner_radius,
                                     self.meta, self.visual)

    @property
    def _outer_region(self):
        return self._component_class(self.center, self.outer_radius,
                                     self.meta, self.visual)

    def to_sky(self, wcs):
        center = wcs.pixel_to_world(self.center.x, self.center.y)
        _, pixscale, _ = pixel_scale_angle_at_skycoord(center, wcs)
        inner_radius = self.inner_radius * u.pix * pixscale
        outer_radius = self.outer_radius * u.pix * pixscale
        return CircleAnnulusSkyRegion(center, inner_radius, outer_radius,
                                      self.meta.copy(), self.visual.copy())


class CircleAnnulusSkyRegion(SkyRegion):
    """
    A circular annulus in sky coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the center of the annulus.
    inner_radius : `~astropy.units.Quantity`
        The inner radius of the annulus in angular units.
    outer_radius : `~astropy.units.Quantity`
        The outer radius of the annulus in angular units.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.
    """

    _params = ('center', 'inner_radius', 'outer_radius')
    center = ScalarSkyCoord('The center position as a |SkyCoord|.')
    inner_radius = PositiveScalarAngle('The inner radius as a |Quantity| '
                                       'angle.')
    outer_radius = PositiveScalarAngle('The outer radius as a |Quantity| '
                                       'angle.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, center, inner_radius, outer_radius, meta=None,
                 visual=None):
        self.center = center
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

        if inner_radius >= outer_radius:
            raise ValueError('outer_radius must be greater than inner_radius')

    def to_pixel(self, wcs):
        center, pixscale, _ = pixel_scale_angle_at_skycoord(self.center, wcs)
        inner_radius = (self.inner_radius / pixscale).to(u.pix).value
        outer_radius = (self.outer_radius / pixscale).to(u.pix).value
        return CircleAnnulusPixelRegion(center, inner_radius, outer_radius,
                                        meta=self.meta.copy(),
                                        visual=self.visual.copy())


class AsymmetricAnnulusPixelRegion(AnnulusPixelRegion):
    """
    Helper class for asymmetric annuli sky regions.

    Used for ellipse and rectangle pixel annulus regions.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the annulus.
    center : `~regions.PixCoord`
        The position of the center of the annulus.
    inner_width : float
        The inner width of the annulus (before rotation) in pixels.
    outer_width : float
        The outer width of the annulus (before rotation) in pixels.
    inner_height : float
        The inner height of the annulus (before rotation) in pixels.
    outer_height : float
        The outer height of the annulus (before rotation) in pixels.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the annulus, measured anti-clockwise. If
        set to zero (the default), the width axis is lined up with the x
        axis.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.
    """

    _params = ('center', 'inner_width', 'outer_width', 'inner_height',
               'outer_height', 'angle')
    center = ScalarPixCoord('The center pixel position as a |PixCoord|.')
    inner_width = PositiveScalar('The inner width (before rotation) in '
                                 'pixels as a float.')
    outer_width = PositiveScalar('The outer width (before rotation) in '
                                 'pixels as a float.')
    inner_height = PositiveScalar('The inner height (before rotation) in '
                                  'pixels as a float.')
    outer_height = PositiveScalar('The outer height (before rotation) in '
                                  'pixels as a float.')
    angle = ScalarAngle('The rotation angle measured anti-clockwise as a '
                        '|Quantity| angle.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, center, inner_width, outer_width, inner_height,
                 outer_height, angle=0 * u.deg, meta=None, visual=None):
        self.center = center
        self.inner_width = inner_width
        self.outer_width = outer_width
        self.inner_height = inner_height
        self.outer_height = outer_height
        self.angle = angle
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

        if inner_width >= outer_width:
            raise ValueError('outer_width must be greater than inner_width')
        if inner_height >= outer_height:
            raise ValueError('outer_height must be greater than inner_height')

    @property
    def _inner_region(self):
        return self._component_class(self.center, self.inner_width,
                                     self.inner_height, self.angle,
                                     self.meta, self.visual)

    @property
    def _outer_region(self):
        return self._component_class(self.center, self.outer_width,
                                     self.outer_height, self.angle,
                                     self.meta, self.visual)

    def to_sky_args(self, wcs):
        center = wcs.pixel_to_world(self.center.x, self.center.y)
        _, pixscale, north_angle = pixel_scale_angle_at_skycoord(center, wcs)
        inner_width = (self.inner_width * u.pix * pixscale).to(u.arcsec)
        outer_width = (self.outer_width * u.pix * pixscale).to(u.arcsec)
        inner_height = (self.inner_height * u.pix * pixscale).to(u.arcsec)
        outer_height = (self.outer_height * u.pix * pixscale).to(u.arcsec)
        # region sky angles are defined relative to the WCS longitude axis;
        # photutils aperture sky angles are defined as the PA of the
        # semimajor axis (i.e., relative to the WCS latitude axis)
        angle = self.angle - (north_angle - 90 * u.deg)

        return (center, inner_width, outer_width, inner_height, outer_height,
                angle)


class AsymmetricAnnulusSkyRegion(SkyRegion):
    """
    Helper class for asymmetric annuli sky regions.

    Used for ellipse and rectangle sky annulus regions.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the center of the annulus.
    inner_width : `~astropy.units.Quantity`
        The inner width of the annulus (before rotation) as an angle.
    outer_width : `~astropy.units.Quantity`
        The outer width of the annulus (before rotation) as an angle.
    inner_height : `~astropy.units.Quantity`
        The inner height of the annulus (before rotation) as an angle.
    outer_height : `~astropy.units.Quantity`
        The outer height of the annulus (before rotation) as an angle.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the annulus, measured anti-clockwise. If
        set to zero (the default), the width axis is lined up with the
        longitude axis of the celestial coordinates.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.
    """

    _params = ('center', 'inner_width', 'outer_width', 'inner_height',
               'outer_height', 'angle')

    center = ScalarSkyCoord('The center position as a |SkyCoord|.')
    inner_width = PositiveScalarAngle('The inner width (before rotation) as '
                                      'a |Quantity| angle.')
    outer_width = PositiveScalarAngle('The outer width (before rotation) as '
                                      'a |Quantity| angle.')
    inner_height = PositiveScalarAngle('The inner height (before rotation) '
                                       'as a |Quantity| angle.')
    outer_height = PositiveScalarAngle('The outer height (before rotation) '
                                       'as a |Quantity| angle.')
    angle = ScalarAngle('The rotation angle measured anti-clockwise as a'
                        '|Quantity| angle.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, center, inner_width, outer_width, inner_height,
                 outer_height, angle=0 * u.deg, meta=None, visual=None):
        self.center = center
        self.inner_width = inner_width
        self.outer_width = outer_width
        self.inner_height = inner_height
        self.outer_height = outer_height
        self.angle = angle
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

        if inner_width >= outer_width:
            raise ValueError('outer_width must be greater than inner_width')
        if inner_height >= outer_height:
            raise ValueError('outer_height must be greater than inner_height')

    def to_pixel_args(self, wcs):
        center, pixscale, north_angle = pixel_scale_angle_at_skycoord(
            self.center, wcs)
        center = PixCoord(center.x, center.y)
        inner_width = (self.inner_width / pixscale).to(u.pix).value
        outer_width = (self.outer_width / pixscale).to(u.pix).value
        inner_height = (self.inner_height / pixscale).to(u.pix).value
        outer_height = (self.outer_height / pixscale).to(u.pix).value
        # region sky angles are defined relative to the WCS longitude axis;
        # photutils aperture sky angles are defined as the PA of the
        # semimajor axis (i.e., relative to the WCS latitude axis)
        angle = self.angle + (north_angle - 90 * u.deg)

        return (center, inner_width, outer_width, inner_height, outer_height,
                angle)


class EllipseAnnulusPixelRegion(AsymmetricAnnulusPixelRegion):
    """
    A elliptical annulus in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the elliptical annulus.
    inner_width : float
        The inner width of the elliptical annulus (before rotation) in
        pixels.
    outer_width : float
        The outer width of the elliptical annulus (before rotation) in
        pixels.
    inner_height : float
        The inner height of the elliptical annulus (before rotation) in
        pixels.
    outer_height : float
        The outer height of the elliptical annulus (before rotation) in
        pixels.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the elliptical annulus, measured
        anti-clockwise. If set to zero (the default), the width axis is
        lined up with the x axis.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from astropy.coordinates import Angle
        from regions import PixCoord, EllipseAnnulusPixelRegion
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1)

        reg = EllipseAnnulusPixelRegion(PixCoord(6, 6),
                                        inner_width=5.5,
                                        outer_width=8.5,
                                        inner_height=3.5,
                                        outer_height=6.5,
                                        angle=Angle('45deg'))
        patch = reg.plot(ax=ax, facecolor='none', edgecolor='red', lw=2,
                         label='Ellipse Annulus')

        ax.legend(handles=(patch,), loc='upper center')
        ax.set_xlim(-5, 20)
        ax.set_ylim(-5, 20)
        ax.set_aspect('equal')
    """

    # duplicated from AsymmetricAnnulusPixelRegion because otherwise Sphinx
    # ignores the docstrings in the parent class
    center = ScalarPixCoord('The center pixel position as a |PixCoord|.')
    inner_width = PositiveScalar('The inner width (before rotation) in '
                                 'pixels as a float.')
    outer_width = PositiveScalar('The outer width (before rotation) in '
                                 'pixels as a float.')
    inner_height = PositiveScalar('The inner height (before rotation) in '
                                  'pixels as a float.')
    outer_height = PositiveScalar('The outer height (before rotation) in '
                                  'pixels as a float.')
    angle = ScalarAngle('The rotation angle measured anti-clockwise as a '
                        '|Quantity| angle.')

    _component_class = EllipsePixelRegion

    def __init__(self, center, inner_width, outer_width, inner_height,
                 outer_height, angle=0 * u.deg, meta=None, visual=None):
        super().__init__(center, inner_width, outer_width, inner_height,
                         outer_height, angle, meta, visual)

    def to_sky(self, wcs):
        return EllipseAnnulusSkyRegion(*self.to_sky_args(wcs),
                                       meta=self.meta.copy(),
                                       visual=self.visual.copy())


class EllipseAnnulusSkyRegion(AsymmetricAnnulusSkyRegion):
    """
    A elliptical annulus in `~astropy.coordinates.SkyCoord` coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the center of the elliptical annulus.
    inner_width : `~astropy.units.Quantity`
        The inner width of the elliptical annulus (before rotation) as
        an angle.
    outer_width : `~astropy.units.Quantity`
        The outer width of the elliptical annulus (before rotation) as
        an angle.
    inner_height : `~astropy.units.Quantity`
        The inner height of the elliptical annulus (before rotation) as
        an angle.
    outer_height : `~astropy.units.Quantity`
        The outer height of the elliptical annulus (before rotation) as
        an angle.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the elliptical annulus, measured
        anti-clockwise. If set to zero (the default), the width axis is
        lined up with the longitude axis of the celestial coordinates.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.
    """

    # duplicated from AsymmetricAnnulusSkyRegion because otherwise Sphinx
    # ignores the docstrings in the parent class
    center = ScalarSkyCoord('The center position as a |SkyCoord|.')
    inner_width = PositiveScalarAngle('The inner width (before rotation) as '
                                      'a |Quantity| angle.')
    outer_width = PositiveScalarAngle('The outer width (before rotation) as '
                                      'a |Quantity| angle.')
    inner_height = PositiveScalarAngle('The inner height (before rotation) '
                                       'as a |Quantity| angle.')
    outer_height = PositiveScalarAngle('The outer height (before rotation) '
                                       'as a |Quantity| angle.')
    angle = ScalarAngle('The rotation angle measured anti-clockwise as a '
                        '|Quantity| angle.')

    _component_class = EllipseSkyRegion

    def __init__(self, center, inner_width, outer_width, inner_height,
                 outer_height, angle=0 * u.deg, meta=None, visual=None):
        super().__init__(center, inner_width, outer_width, inner_height,
                         outer_height, angle, meta, visual)

    def to_pixel(self, wcs):
        return EllipseAnnulusPixelRegion(*self.to_pixel_args(wcs),
                                         meta=self.meta.copy(),
                                         visual=self.visual.copy())


class RectangleAnnulusPixelRegion(AsymmetricAnnulusPixelRegion):
    """
    A rectangular annulus in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the rectangular annulus.
    inner_width : float
        The inner width of the rectangular annulus (before rotation) in
        pixels.
    outer_width : float
        The outer width of the rectangular annulus (before rotation) in
        pixels.
    inner_height : float
        The inner height of the rectangular annulus (before rotation) in
        pixels.
    outer_height : float
        The outer height of the rectangular annulus (before rotation) in
        pixels.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the rectangular annulus, measured
        anti-clockwise. If set to zero (the default), the width axis is
        lined up with the x axis.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from astropy.coordinates import Angle
        from regions import PixCoord, RectangleAnnulusPixelRegion
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1)

        reg = RectangleAnnulusPixelRegion(PixCoord(x=6, y=6),
                                          inner_width=5.5,
                                          outer_width=8.5,
                                          inner_height=3.5,
                                          outer_height=6.5,
                                          angle=Angle('45deg'))
        patch = reg.plot(ax=ax, facecolor='none', edgecolor='red', lw=2,
                         label='Rectangle Annulus')

        ax.legend(handles=(patch,), loc='upper center')
        ax.set_xlim(-5, 20)
        ax.set_ylim(-5, 20)
        ax.set_aspect('equal')
    """

    # duplicated from AsymmetricAnnulusPixelRegion because otherwise Sphinx
    # ignores the docstrings in the parent class
    center = ScalarPixCoord('The center pixel position as a |PixCoord|.')
    inner_width = PositiveScalar('The inner width (before rotation) in '
                                 'pixels as a float.')
    outer_width = PositiveScalar('The outer width (before rotation) in '
                                 'pixels as a float.')
    inner_height = PositiveScalar('The inner height (before rotation) in '
                                  'pixels as a float.')
    outer_height = PositiveScalar('The outer height (before rotation) in '
                                  'pixels as a float.')
    angle = ScalarAngle('The rotation angle measured anti-clockwise as a '
                        '|Quantity| angle.')

    _component_class = RectanglePixelRegion

    def __init__(self, center, inner_width, outer_width, inner_height,
                 outer_height, angle=0 * u.deg, meta=None, visual=None):
        super().__init__(center, inner_width, outer_width, inner_height,
                         outer_height, angle, meta, visual)

    def to_sky(self, wcs):
        return RectangleAnnulusSkyRegion(*self.to_sky_args(wcs),
                                         meta=self.meta.copy(),
                                         visual=self.visual.copy())


class RectangleAnnulusSkyRegion(AsymmetricAnnulusSkyRegion):
    """
    A rectangular annulus in `~astropy.coordinates.SkyCoord`
    coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the center of the rectangular annulus.
    inner_width : `~astropy.units.Quantity`
        The inner width of the rectangular annulus (before rotation) as
        an angle.
    outer_width : `~astropy.units.Quantity`
        The outer width of the rectangular annulus (before rotation) as
        an angle.
    inner_height : `~astropy.units.Quantity`
        The inner height of the rectangular annulus (before rotation) as
        an angle.
    outer_height : `~astropy.units.Quantity`
        The outer height of the rectangular annulus (before rotation) as
        an angle.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the rectangular annulus, measured
        anti-clockwise. If set to zero (the default), the width axis is
        lined up with the longitude axis of the celestial coordinates.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.
    """

    # duplicated from AsymmetricAnnulusSkyRegion because otherwise Sphinx
    # ignores the docstrings in the parent class
    center = ScalarSkyCoord('The center position as a |SkyCoord|.')
    inner_width = PositiveScalarAngle('The inner width (before rotation) as '
                                      'a |Quantity| angle.')
    outer_width = PositiveScalarAngle('The outer width (before rotation) as '
                                      'a |Quantity| angle.')
    inner_height = PositiveScalarAngle('The inner height (before rotation) '
                                       'as a |Quantity| angle.')
    outer_height = PositiveScalarAngle('The outer height (before rotation) '
                                       'as a |Quantity| angle.')
    angle = ScalarAngle('The rotation angle measured anti-clockwise as a '
                        '|Quantity| angle.')

    _component_class = RectangleSkyRegion

    def __init__(self, center, inner_width, outer_width, inner_height,
                 outer_height, angle=0 * u.deg, meta=None, visual=None):
        super().__init__(center, inner_width, outer_width, inner_height,
                         outer_height, angle, meta, visual)

    def to_pixel(self, wcs):
        return RectangleAnnulusPixelRegion(*self.to_pixel_args(wcs),
                                           meta=self.meta.copy(),
                                           visual=self.visual.copy())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Regions is Astropy coordinated package to provide tools for region
handling.
"""

from ._utils.examples import *  # noqa: F401, F403
from .core import *  # noqa: F401, F403
from .io import *  # noqa: F401, F403
from .shapes import *  # noqa: F401, F403

try:
    from .version import version as __version__
except ImportError:
    __version__ = ''


# Set the bibtex entry to the article referenced in CITATION.rst.
def _get_bibtex():
    import os
    citation_file = os.path.join(os.path.dirname(__file__), 'CITATION.rst')

    with open(citation_file) as citation:
        refs = citation.read().split('@software')[1:]
        if len(refs) == 0:
            return ''
        bibtexreference = f'@software{refs[0]}'
    return bibtexreference


__citation__ = __bibtex__ = _get_bibtex()

del _get_bibtex



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the fits subpackage.
"""

import warnings

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.table import QTable
from astropy.utils.data import get_pkg_data_filenames
from astropy.utils.exceptions import AstropyUserWarning
from numpy.testing import assert_equal

from regions.core import PixCoord, RegionMeta, Regions
from regions.io.fits.core import FITSParserError
from regions.shapes import (CirclePixelRegion, CircleSkyRegion,
                            LinePixelRegion, RectangleAnnulusPixelRegion,
                            TextPixelRegion)
from regions.tests.helpers import assert_region_allclose


def test_roundtrip(tmpdir):
    filenames = get_pkg_data_filenames('data', pattern='*.fits')

    # AstropyUserWarning will be emitted only for some of the files
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', AstropyUserWarning)
        # Check that all test files are readable
        for filename in filenames:
            regions = Regions.read(filename, format='fits')
            assert len(regions) > 0

        tempfile = tmpdir.join('tmp.fits').strpath
        regions.write(tempfile, format='fits', overwrite=True)
        regions2 = Regions.read(tempfile, format='fits')
        for reg1, reg2 in zip(regions, regions2, strict=True):
            assert_region_allclose(reg1, reg2)


def test_only_pixel_regions():
    reg_sky = CircleSkyRegion(SkyCoord(1, 2, unit='deg'), 5 * u.deg)
    reg = Regions([reg_sky])

    match = 'Sky regions cannot be serialized'
    with pytest.warns(AstropyUserWarning, match=match):
        result = reg.serialize(format='fits')
        assert len(result) == 0

    reg_pix = CirclePixelRegion(PixCoord(10, 10), 5)
    reg = Regions([reg_sky, reg_pix])
    match = 'Sky regions cannot be serialized'
    with pytest.warns(AstropyUserWarning, match=match):
        result = reg.serialize(format='fits')
        assert len(result) == 1


def test_valid_columns():
    t = QTable([[1, 2, 3]], names=('a'))
    with pytest.raises(FITSParserError) as excinfo:
        Regions.parse(t, format='fits')
        estr = "'a' is not a valid column name"
        assert estr in str(excinfo.value)


def test_invalid_regions():
    center = PixCoord(42, 43)
    region1 = RectangleAnnulusPixelRegion(center, 4.2, 5.2, 7.2, 8.2,
                                          angle=15 * u.deg)
    region2 = LinePixelRegion(start=center, end=PixCoord(x=52, y=53))
    region3 = TextPixelRegion(center, text='Example Text')
    regions = (region1, region2, region3)

    match = 'cannot be serialized using the FITS format'
    for reg in regions:
        with pytest.warns(AstropyUserWarning, match=match):
            _ = reg.serialize(format='fits')


def test_valid_row():
    x = [1] * u.pix
    y = [2] * u.pix
    shapes = ['CIRCLE']
    tbl = QTable([x, y, shapes], names=('X', 'Y', 'SHAPE'))
    match = 'Table columns are missing'
    with pytest.warns(AstropyUserWarning, match=match):
        Regions.parse(tbl, format='fits')

    # test invalid shape
    shapes = ['INVALID']
    tbl2 = QTable([x, y, shapes], names=('X', 'Y', 'SHAPE'))
    with pytest.raises(FITSParserError) as excinfo:
        Regions.parse(tbl2, format='fits')
        estr = "'invalid' is not a valid FITS region shape"
        assert estr in str(excinfo.value)

    shapes = ['PIE']
    rotang = [[20, 30]] * u.deg
    tbl3 = QTable([x, y, shapes, rotang], names=('X', 'Y', 'SHAPE', 'ROTANG'))
    match = "'pie' is not supported"
    with pytest.warns(AstropyUserWarning, match=match):
        Regions.parse(tbl3, format='fits')


def test_components():
    center = PixCoord(10, 10)
    regions = Regions((CirclePixelRegion(center, 3),
                       CirclePixelRegion(center, 5),
                       CirclePixelRegion(center, 7),
                       CirclePixelRegion(center, 9)))
    tbl1 = regions.serialize(format='fits')
    assert 'COMPONENT' not in tbl1.colnames

    components = [None, None, None, None]
    for region, component in zip(regions, components, strict=True):
        region.meta = RegionMeta({'component': component})
    tbl2 = regions.serialize(format='fits')
    assert 'COMPONENT' not in tbl2.colnames

    components = np.arange(4)
    for region, component in zip(regions, components, strict=True):
        region.meta = RegionMeta({'component': component})
    tbl3 = regions.serialize(format='fits')
    assert 'COMPONENT' in tbl3.colnames
    assert_equal(tbl3['COMPONENT'], components)

    components = [1, 2, None, 4]
    for region, component in zip(regions, components, strict=True):
        region.meta = RegionMeta({'component': component})
    tbl4 = regions.serialize(format='fits')
    assert 'COMPONENT' in tbl4.colnames
    assert_equal(tbl4['COMPONENT'], [1, 2, 5, 4])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the regions module.
"""

import pytest
from astropy.table import Table

from regions.core import PixCoord, Regions
from regions.shapes import CirclePixelRegion


def test_regions_inputs():
    regs = []
    for radius in range(1, 5):
        center = PixCoord(14, 21)
        regs.append(CirclePixelRegion(center, radius=radius))

    regions = Regions(regs)
    assert len(regions) == 4


def test_regions_no_input():
    regions = Regions()
    assert len(regions) == 0

    regions = Regions([])
    assert len(regions) == 0


def test_regions_invalid_input():
    match = "'int' object is not iterable"
    with pytest.raises(TypeError, match=match):
        Regions(1)

    match = 'Input regions must be a list of Region objects'
    with pytest.raises(TypeError, match=match):
        Regions([1])
    with pytest.raises(TypeError, match=match):
        Regions([1, 2, 3])


def test_regions_append():
    regions = Regions()

    for radius in range(1, 5):
        center = PixCoord(14, 21)
        regions.append(CirclePixelRegion(center, radius=radius))
    assert len(regions) == 4

    match = 'Input region must be a Region object'
    with pytest.raises(TypeError, match=match):
        regions.append(1)


def test_regions_extend():
    regions = Regions()

    regs = []
    for radius in range(1, 5):
        center = PixCoord(14, 21)
        regs.append(CirclePixelRegion(center, radius=radius))

    regions.extend(regs)
    assert len(regions) == 4

    regions.extend(regions)
    assert len(regions) == 8

    match = 'Input regions must be a list of Region objects'
    with pytest.raises(TypeError, match=match):
        regions.extend([1])


def test_regions_methods():
    regions = Regions()

    for radius in range(1, 5):
        center = PixCoord(14, 21)
        regions.append(CirclePixelRegion(center, radius=radius))

    reg_slc = regions[0:2]
    assert len(reg_slc) == 2

    reg = CirclePixelRegion(PixCoord(0, 0), radius=1)
    regions.insert(0, reg)
    assert len(regions) == 5
    assert regions[0] == reg

    regions2 = regions.copy()
    regions.reverse()
    assert regions.regions == regions2.regions[::-1]

    outreg = regions.pop(-1)
    assert outreg == reg


def test_regions_get_formats():
    reg = CirclePixelRegion(PixCoord(0, 0), radius=1)
    regions = Regions([reg])
    tbl = regions.get_formats()
    assert isinstance(tbl, Table)
    assert len(tbl) == 3


def test_regions_repr():
    reg = CirclePixelRegion(PixCoord(0, 0), radius=1)
    regions = Regions([reg])
    regions_str = '[<CirclePixelRegion(center=PixCoord(x=0, y=0), radius=1)>]'
    regions_repr = f'<Regions({regions_str})>'
    assert str(regions) == regions_str
    assert repr(regions) == regions_repr
